# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2020 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
import ast
import typing as typ
import logging

from . import common
from . import checker_base as cb

# TODO (mb 2020-05-28):
#   instead of functools.singledispatch
#   from singledispatch import singledispatch
#   https://pypi.org/project/singledispatch/
#
#   instead of functools.lru_cache
#   from backports import functools_lru_cache
#   https://pypi.org/project/backports.functools-lru-cache/
#

logger = logging.getLogger(__name__)


class ModuleVersionInfo(typ.NamedTuple):

    available_since : str
    backport_module : typ.Optional[str]
    backport_package: typ.Optional[str]


MAYBE_UNUSABLE_MODULES = {
    # case 1 (simple). Always error because no backport available
    'asyncio': ModuleVersionInfo("3.4", None, None),
    'zipapp' : ModuleVersionInfo("3.5", None, None),
    # case 2 (simple). Always error because modules have different names and only
    #   backport should be used
    'csv'                : ModuleVersionInfo("3.0", "backports.csv"      , "backports.csv"),
    'selectors'          : ModuleVersionInfo("3.4", "selectors2"         , "selectors2"),
    'pathlib'            : ModuleVersionInfo("3.4", "pathlib2"           , "pathlib2"),
    "importlib.resources": ModuleVersionInfo("3.7", "importlib_resources", "importlib_resources"),
    'inspect'            : ModuleVersionInfo("3.6", "inspect2"           , "inspect2"),
    # case 3 (hard). Modules have the same name.
    # - By default, only logger.warning if these are modules are imported.
    # - If opt-in via '--install-requires' option or
    #   'install_requires' argument of 'lib3to6.fix', check that they
    #   have been explicitly whitelisted.
    'lzma'       : ModuleVersionInfo("3.3", "lzma"       , "backports.lzma"),
    'ipaddress'  : ModuleVersionInfo("3.4", "ipaddress"  , "py2-ipaddress"),
    'enum'       : ModuleVersionInfo("3.4", "enum"       , "enum34"),
    'typing'     : ModuleVersionInfo("3.5", 'typing'     , 'typing'),
    'secrets'    : ModuleVersionInfo("3.6", "secrets"    , "python2-secrets"),
    'statistics' : ModuleVersionInfo("3.4", "statistics" , "statistics"),
    'dataclasses': ModuleVersionInfo("3.7", "dataclasses", "dataclasses"),
    'contextvars': ModuleVersionInfo("3.7", "contextvars", "contextvars"),
}


def _iter_module_names(node: ast.AST) -> typ.Iterable[str]:
    if isinstance(node, ast.Import):
        for alias in node.names:
            yield alias.name
    elif isinstance(node, ast.ImportFrom):
        mname = node.module
        if mname:
            yield mname


def _iter_maybe_unusable_modules(node: ast.AST) -> typ.Iterable[typ.Tuple[str, ModuleVersionInfo]]:
    for mname in _iter_module_names(node):
        vnfo = MAYBE_UNUSABLE_MODULES.get(mname)
        if vnfo:
            yield (mname, vnfo)


class NoUnusableImportsChecker(cb.CheckerBase):
    # NOTE (mb 2020-05-28): The (apparent double negation) naming of this does
    #   make sense; the name "OnlyUsableImportsChecker" would not be better,
    #   because this doesn't check all imports ever in existence, it only checks
    #   that unusable imports are not used at the top level.

    # NOTE (mb 2020-05-28): This checker only checks top level imports.
    #   This allows for the common idom to work without errors.
    #
    #   try:
    #       import newmodule
    #   except ImportError:
    #       improt backport_module as newmodule

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> None:
        # NOTE (mb 2020-05-28):
        #   - Strict mode fails hard
        #   - Only raises an error if an unusable module is not used
        #   - Only warns about backported modules after
        #     opt-in to this check (by using the fix(install_requires) parameter).
        #   - Existing systems will have to update
        #     their config for this check to work and we don't want to
        #     break them.

        install_requires = ctx.cfg.install_requires

        target_version = ctx.cfg.target_version
        for node in ast.iter_child_nodes(tree):
            for mname, vnfo in _iter_maybe_unusable_modules(node):
                if target_version >= vnfo.available_since:
                    # target supports the newer name
                    continue

                bppkg = vnfo.backport_package

                # if the backport has a different name, then there is no
                # excuse not to use it -> hard error
                is_backport_name_same = mname == vnfo.backport_module

                is_whitelisted = (
                    is_backport_name_same
                    and install_requires is not None
                    and bppkg in install_requires
                )
                if is_whitelisted:
                    continue

                # From here, we either error or at least show a warning.

                # if there is no backport, then the import can obviously only
                # be using the stdlib module -> hard error
                is_backported  = vnfo.backport_package is not None
                is_strict_mode = install_requires      is not None
                is_hard_error  = not is_backported or is_strict_mode or not is_backport_name_same

                vnfo_msg = (
                    f"This module is available since Python {vnfo.available_since}, "
                    f"but you configured target_version='{target_version}'."
                )

                if is_hard_error:
                    errmsg = f"Prohibited import '{mname}'. {vnfo_msg}"
                    if bppkg:
                        errmsg += f" Use 'https://pypi.org/project/{bppkg}' instead."
                    else:
                        errmsg += " No backported for this package is known."

                    raise common.CheckError(errmsg, node)
                else:
                    lineno = common.get_node_lineno(node)
                    loc    = f"{ctx.filepath}@{lineno}"
                    logger.warning(f"{loc}: Use of import '{mname}'. {vnfo_msg}")



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import ast
import typing as typ

from . import common

# NOTE (mb 2018-06-24): Version info pulled from:
# https://docs.python.org/3/library/__future__.html


class FixerBase:

    version_info       : common.VersionInfo
    required_imports   : typ.Set[common.ImportDecl]
    module_declarations: typ.Set[str]

    def __init__(self) -> None:
        self.required_imports    = set()
        self.module_declarations = set()

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> ast.Module:
        raise NotImplementedError()


class TransformerFixerBase(FixerBase, ast.NodeTransformer):
    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> ast.Module:
        try:
            new_tree = self.visit(tree)
            return typ.cast(ast.Module, new_tree)
        except common.FixerError as ex:
            if ex.module is None:
                ex.module = tree
            raise



#!/usr/bin/env python
# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import io
import re
import sys
import typing as typ
import difflib
import logging

import click

from . import common
from . import packaging
from . import transpile

try:
    import pretty_traceback

    pretty_traceback.install(envvar='ENABLE_PRETTY_TRACEBACK')
except ImportError:
    pass  # no need to fail because of missing dev dependency


logger = logging.getLogger("lib3to6")


def _configure_logging(verbose: int = 0) -> None:
    if verbose >= 2:
        log_format = "%(asctime)s.%(msecs)03d %(levelname)-7s %(name)-17s - %(message)s"
        log_level  = logging.DEBUG
    elif verbose == 1:
        log_format = "%(levelname)-7s - %(message)s"
        log_level  = logging.INFO
    else:
        log_format = "%(levelname)-7s - %(message)s"
        log_level  = logging.INFO

    logging.basicConfig(level=log_level, format=log_format, datefmt="%Y-%m-%dT%H:%M:%S")
    logger.debug("Logging configured.")


click.disable_unicode_literals_warning = True


def _print_diff(source_text: str, fixed_source_text: str) -> None:
    differ = difflib.Differ()

    source_lines       = source_text.splitlines()
    fixed_source_lines = fixed_source_text.splitlines()
    diff_lines         = differ.compare(source_lines, fixed_source_lines)
    if not sys.stdout.isatty():
        click.echo("\n".join(diff_lines))
        return

    for line in diff_lines:
        if line.startswith("+ "):
            click.echo("\u001b[32m" + line + "\u001b[0m")
        elif line.startswith("- "):
            click.echo("\u001b[31m" + line + "\u001b[0m")
        elif line.startswith("? "):
            click.echo("\u001b[36m" + line + "\u001b[0m")
        else:
            click.echo(line)
    print()


__INSTALL_REQUIRES_HELP = """
install_requires package dependencies (space separated).
Functions as a whitelist for backported modules.
"""

__DEFAULT_MODE_HELP = """
[enabled/disabled] Default transpile mode.
To transpile some files but not others.
"""


@click.command()
@click.option(
    '-v',
    '--verbose',
    count=True,
    help="Control log level. -vv for debug level.",
)
@click.option(
    "--target-version",
    default="2.7",
    metavar="<version>",
    help="Target version of python.",
)
@click.option(
    "--diff",
    default=False,
    is_flag=True,
    help="Output diff instead of transpiled source.",
)
@click.option(
    "--in-place",
    default=False,
    is_flag=True,
    help="Write result back to input file.",
)
@click.option(
    "--install-requires",
    default=None,
    metavar="<packages>",
    help=__INSTALL_REQUIRES_HELP.strip(),
)
@click.option(
    "--default-mode",
    default='enabled',
    metavar="<mode>",
    help=__DEFAULT_MODE_HELP.strip(),
)
@click.argument(
    "source_files",
    metavar="<source_file>",
    nargs=-1,
    type=click.File(mode="r"),
)
def main(
    target_version  : str,
    diff            : bool,
    in_place        : bool,
    install_requires: typ.Optional[str],
    source_files    : typ.Sequence[io.TextIOWrapper],
    default_mode    : str = 'enabled',
    verbose         : int = 0,
) -> None:
    _configure_logging(verbose)

    has_opt_error = False

    if target_version and not re.match(r"[0-9]+\.[0-9]+", target_version):
        print(f"Invalid argument --target-version={target_version}")
        has_opt_error = True

    if default_mode not in ('enabled', 'disabled'):
        print(f"Invalid argument --default-mode={default_mode}")
        print("    Must be either 'enabled' or 'disabled'")
        has_opt_error = True

    if not any(source_files):
        print("No files.")
        has_opt_error = True

    if has_opt_error:
        sys.exit(1)

    cfg = packaging.eval_build_config(
        target_version=target_version,
        install_requires=install_requires,
        default_mode=default_mode,
    )
    for src_file in source_files:
        ctx         = common.BuildContext(cfg, src_file.name)
        source_text = src_file.read()
        try:
            fixed_source_text = transpile.transpile_module(ctx, source_text)
        except common.CheckError as err:
            loc = src_file.name
            if err.lineno >= 0:
                loc += "@" + str(err.lineno)

            err.args = (loc + " - " + err.args[0],) + err.args[1:]
            raise

        if diff:
            _print_diff(source_text, fixed_source_text)
        elif in_place:
            with io.open(src_file.name, mode="w") as fobj:
                fobj.write(fixed_source_text)
        else:
            print(fixed_source_text)


if __name__ == '__main__':
    # NOTE (mb 2020-07-18): click supplies the parameters
    # pylint:disable=no-value-for-parameter
    main()  # type: ignore



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import ast
import typing as typ
import builtins

PackageName      = str
PackageDirectory = str
PackageDir       = typ.Dict[PackageName, PackageDirectory]

InstallRequires = typ.Optional[typ.Set[str]]


ConstantNodeTypes: typ.Tuple[typ.Type[ast.Constant], ...] = (ast.Constant,)

# Deprecated since version 3.8: Class ast.Constant is now used for all
# constants. Old classes ast.Num, ast.Str, ast.Bytes, ast.NameConstant and
# ast.Ellipsis are still available, but they will be removed in future Python
# releases.

if hasattr(ast, 'Num'):
    ConstantNodeTypes += (ast.Num, ast.Str, ast.Bytes, ast.NameConstant, ast.Ellipsis)


LeafNodeTypes = ConstantNodeTypes + (
    ast.Name,
    ast.cmpop,
    ast.boolop,
    ast.operator,
    ast.unaryop,
    ast.expr_context,
)


ContainerNodes = (ast.List, ast.Set, ast.Tuple)


class BuildConfig(typ.NamedTuple):

    target_version  : str  # e.g. "2.7"
    cache_enabled   : bool
    default_mode    : str
    fixers          : str
    checkers        : str
    install_requires: InstallRequires


class BuildContext(typ.NamedTuple):

    cfg     : BuildConfig
    filepath: str


def init_build_context(
    target_version  : str             = "2.7",
    cache_enabled   : bool            = True,
    default_mode    : str             = 'enabled',
    fixers          : str             = "",
    checkers        : str             = "",
    install_requires: InstallRequires = None,
    filepath        : str             = "<filepath>",
) -> BuildContext:
    cfg = BuildConfig(
        target_version=target_version,
        cache_enabled=cache_enabled,
        default_mode=default_mode,
        fixers=fixers,
        checkers=checkers,
        install_requires=install_requires,
    )
    return BuildContext(cfg=cfg, filepath=filepath)


# Additional items:
#   'filepath': "/path/to/inputfile.py"


class InvalidPackage(Exception):
    pass


def get_node_lineno(node: ast.AST = None, parent: ast.AST = None) -> int:
    if isinstance(node, (ast.stmt, ast.expr)):
        return node.lineno
    if isinstance(parent, (ast.stmt, ast.expr)):
        return parent.lineno
    return -1


class CheckError(Exception):

    lineno: int

    def __init__(self, msg: str, node: ast.AST = None, parent: ast.AST = None) -> None:
        self.lineno = get_node_lineno(node, parent)
        super().__init__(msg)


class FixerError(Exception):

    msg   : str
    node  : ast.AST
    module: typ.Optional[ast.Module]

    def __init__(self, msg: str, node: ast.AST, module: ast.Module = None) -> None:
        self.msg    = msg
        self.node   = node
        self.module = module
        super().__init__(msg)


class VersionInfo:
    """Compatibility info for Fixer and Checker classes.

    Used as a class attribute (not an instance attribute) of a fixer or checker.
    The compatability info relates to the fixer/checker rather than the feature
    it deals with. The question that is being ansered is: "should this
    fixer/checker be exectued for this (src_version, tgt_version) pair"
    """

    # since/until is inclusive
    apply_since: typ.List[int]
    apply_until: typ.Optional[typ.List[int]]
    works_since: typ.List[int]
    works_until: typ.Optional[typ.List[int]]

    def __init__(
        self,
        apply_since: str = "1.0",
        apply_until: typ.Optional[str] = None,
        works_since: typ.Optional[str] = None,
        works_until: typ.Optional[str] = None,
    ) -> None:

        self.apply_since = [int(part) for part in apply_since.split(".")]
        if apply_until is None:
            self.apply_until = None
        else:
            self.apply_until = [int(part) for part in apply_until.split(".")]

        if works_since is None:
            # Implicitly, if it's applied since a version, it
            # also works since then.
            self.works_since = self.apply_since
        else:
            self.works_since = [int(part) for part in works_since.split(".")]

        if works_until is None:
            self.works_until = None
        else:
            self.works_until = [int(part) for part in works_until.split(".")]

    def is_required_for(self, version: str) -> bool:
        version_num = [int(part) for part in version.split(".")]
        apply_until = self.apply_until
        if apply_until and apply_until < version_num:
            return False
        return self.apply_since <= version_num

    def is_compatible_with(self, version: str) -> bool:
        version_num = [int(part) for part in version.split(".")]
        works_since = self.works_since
        works_until = self.works_until
        if works_since and version_num < works_since:
            return False
        if works_until and works_until < version_num:
            return False
        return True

    def is_applicable_to(self, source_version: str, target_version: str) -> bool:
        return self.is_required_for(target_version) and self.is_compatible_with(source_version)


# NOTE (mb 2018-06-29): None of the fixers use asname. If a
#   module already has an import using asname, it won't be
#   detected as already imported, and another import (without the
#   asname) will be added to the module.
class ImportDecl(typ.NamedTuple):
    module_name    : str
    import_name    : typ.Optional[str]
    py2_module_name: typ.Optional[str]


# NOTE (mb 2018-06-24): This also includes builtins from py27
#   because a fixer might add a reference to it when building for
#   py27, in which case, we don't want other code to have
#   replaced it with something other than the old builtin.
#   For example replace range -> xrange, so it would be nice if
#   xrange isn't a string or something.
BUILTIN_NAMES = {
    "ArithmeticError",
    "AssertionError",
    "AttributeError",
    "BaseException",
    "BlockingIOError",
    "BrokenPipeError",
    "BufferError",
    "BytesWarning",
    "ChildProcessError",
    "ConnectionAbortedError",
    "ConnectionError",
    "ConnectionRefusedError",
    "ConnectionResetError",
    "DeprecationWarning",
    "EOFError",
    "Ellipsis",
    "EnvironmentError",
    "Exception",
    "False",
    "FileExistsError",
    "FileNotFoundError",
    "FloatingPointError",
    "FutureWarning",
    "GeneratorExit",
    "IOError",
    "ImportError",
    "ImportWarning",
    "IndentationError",
    "IndexError",
    "InterruptedError",
    "IsADirectoryError",
    "KeyError",
    "KeyboardInterrupt",
    "LookupError",
    "MemoryError",
    "ModuleNotFoundError",
    "NameError",
    "None",
    "NotADirectoryError",
    "NotImplemented",
    "NotImplementedError",
    "OSError",
    "OverflowError",
    "PendingDeprecationWarning",
    "PermissionError",
    "ProcessLookupError",
    "RecursionError",
    "ReferenceError",
    "ResourceWarning",
    "RuntimeError",
    "RuntimeWarning",
    "StopAsyncIteration",
    "StopIteration",
    "SyntaxError",
    "SyntaxWarning",
    "SystemError",
    "SystemExit",
    "TabError",
    "TimeoutError",
    "True",
    "TypeError",
    "UnboundLocalError",
    "UnicodeDecodeError",
    "UnicodeEncodeError",
    "UnicodeError",
    "UnicodeTranslateError",
    "UnicodeWarning",
    "UserWarning",
    "ValueError",
    "Warning",
    "ZeroDivisionError",
    "abs",
    "all",
    "any",
    "ascii",
    "bin",
    "bool",
    "bytearray",
    "bytes",
    "callable",
    "chr",
    "classmethod",
    "compile",
    "complex",
    "copyright",
    "credits",
    "delattr",
    "dict",
    "dir",
    "display",
    "divmod",
    "enumerate",
    "eval",
    "exec",
    "filter",
    "float",
    "format",
    "frozenset",
    "get_ipython",
    "getattr",
    "globals",
    "hasattr",
    "hash",
    "help",
    "hex",
    "id",
    "input",
    "int",
    "isinstance",
    "issubclass",
    "iter",
    "len",
    "license",
    "list",
    "locals",
    "map",
    "max",
    "memoryview",
    "min",
    "next",
    "object",
    "oct",
    "open",
    "ord",
    "pow",
    "print",
    "property",
    "range",
    "repr",
    "reversed",
    "round",
    "set",
    "setattr",
    "slice",
    "sorted",
    "staticmethod",
    "str",
    "sum",
    "super",
    "tuple",
    "type",
    "vars",
    "zip",
    "StandardError",
    "apply",
    "basestring",
    "buffer",
    "cmp",
    "coerce",
    "dreload",
    "execfile",
    "file",
    "intern",
    "long",
    "raw_input",
    "reduce",
    "reload",
    "unichr",
    "unicode",
    "xrange",
}

# In case somebody is working on py4k or something

BUILTIN_NAMES.update([name for name in dir(builtins) if not name.startswith("__")])



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import ast
import typing as typ

from . import common
from . import fixer_base as fb


class NamedExprFixer(fb.TransformerFixerBase):

    version_info = common.VersionInfo(apply_since="2.7", apply_until="3.7")

    def _extract_and_replace_named_exprs(
        self, expr: ast.expr
    ) -> typ.Tuple[typ.List[ast.stmt], ast.expr]:
        new_assigns: typ.List[ast.stmt] = []
        if isinstance(expr, ast.NamedExpr):
            new_assigns.append(ast.Assign(targets=[expr.target], value=expr.value))
            new_expr = ast.Name(id=expr.target.id, ctx=ast.Load())
            return new_assigns, new_expr
        else:
            if isinstance(expr, ast.UnaryOp):
                new_sub_assigns, new_operand = self._extract_and_replace_named_exprs(expr.operand)
                new_assigns.extend(new_sub_assigns)
                expr.operand = new_operand
            if isinstance(expr, (ast.BinOp, ast.Compare)):
                new_sub_assigns, new_left = self._extract_and_replace_named_exprs(expr.left)
                new_assigns.extend(new_sub_assigns)
                expr.left = new_left
            if isinstance(expr, ast.BinOp):
                new_sub_assigns, new_right = self._extract_and_replace_named_exprs(expr.right)
                new_assigns.extend(new_sub_assigns)
                expr.right = new_right
            if isinstance(expr, ast.BoolOp):
                new_values = []
                for comp in expr.values:
                    new_sub_assigns, new_comp = self._extract_and_replace_named_exprs(comp)
                    new_values.append(new_comp)
                    new_assigns.extend(new_sub_assigns)
                expr.values = new_values
            if isinstance(expr, ast.Compare):
                new_comparators = []
                for comp in expr.comparators:
                    new_sub_assigns, new_comp = self._extract_and_replace_named_exprs(comp)
                    new_comparators.append(new_comp)
                    new_assigns.extend(new_sub_assigns)
                expr.comparators = new_comparators

            return new_assigns, expr

    def _update(self, nodelist: typ.List[ast.stmt], indent: int = 0) -> None:
        i = 0
        while i < len(nodelist):
            node = nodelist[i]
            if isinstance(node, (ast.If, ast.While)):
                new_assigns, new_test = self._extract_and_replace_named_exprs(node.test)
                if new_assigns and isinstance(node, ast.While):
                    loopcond_name = '__loop_condition'
                    # __loop_condition = True
                    loopcond_init_node = ast.Assign(
                        targets=[ast.Name(id=loopcond_name, ctx=ast.Store())],
                        value=ast.NameConstant(value=True, kind=None),
                    )
                    nodelist.insert(i, loopcond_init_node)

                    # while __loop_condition:
                    node.test = ast.Name(id=loopcond_name, ctx=ast.Load())

                    #   __loop_condition = test
                    loopcond_assign_node = ast.Assign(
                        targets=[ast.Name(id=loopcond_name, ctx=ast.Store())],
                        value=new_test,
                    )
                    # if __loop_condition:
                    new_ifnode = ast.If(
                        test=ast.Name(id=loopcond_name, ctx=ast.Load()),
                        body=node.body,
                        orelse=[],
                    )
                    node.body = new_assigns + [loopcond_assign_node, new_ifnode]
                    i += 1
                else:
                    node.test = new_test
                    if new_assigns:
                        nodelist[i:i] = new_assigns
                    i += 1 + len(new_assigns)
            else:
                i += 1

            for nodelist_name in ('body', 'orelse', 'handlers', 'finalbody'):
                sub_nodelist = getattr(node, nodelist_name, None)
                if sub_nodelist:
                    self._update(sub_nodelist, indent + 1)

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> ast.Module:
        self._update(tree.body)
        return tree



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2020 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
import ast
import typing as typ

from . import common
from . import fixer_base as fb

AstStr = getattr(ast, 'Str', ast.Constant)


ArgUnpackNodes   = (ast.Call, ast.List, ast.Tuple, ast.Set)
KwArgUnpackNodes = (ast.Call, ast.Dict)


def _is_dict_call(node: ast.expr) -> bool:
    return isinstance(node, ast.Call) and isinstance(node.func, ast.Name) and node.func.id == "dict"


def _has_stararg_g12n(node: ast.expr) -> bool:
    if isinstance(node, ast.Call):
        elts = node.args
    elif isinstance(node, (ast.List, ast.Tuple, ast.Set)):
        elts = node.elts
    else:
        raise TypeError(f"Unexpected node: {node}")

    has_starred_arg = False
    for arg in elts:
        # Anything after * means we have to apply the fix
        if has_starred_arg:
            return True
        has_starred_arg = isinstance(arg, ast.Starred)
    return False


def _has_starstarargs_g12n(node: ast.expr) -> bool:
    if isinstance(node, ast.Call):
        has_kwstarred_arg = False
        for keyword in node.keywords:
            if has_kwstarred_arg:
                # Anything after ** means we have to apply the fix
                return True
            has_kwstarred_arg = keyword.arg is None
        return False
    elif isinstance(node, ast.Dict):
        has_kwstarred_arg = False
        for key in node.keys:
            if has_kwstarred_arg:
                # Anything after ** means we have to apply the fix
                return True
            has_kwstarred_arg = key is None
        return False
    else:
        raise TypeError(f"Unexpected node: {node}")


def _node_with_elts(node: ast.AST, new_elts: typ.List[ast.expr]) -> ast.expr:
    if isinstance(node, ast.Call):
        node.args = new_elts
        return node
    elif isinstance(node, ast.List):
        return ast.List(elts=new_elts)
    elif isinstance(node, ast.Set):
        return ast.Set(elts=new_elts)
    elif isinstance(node, ast.Tuple):
        return ast.Tuple(elts=new_elts)
    else:
        raise TypeError(f"Unexpected node type {type(node)}")


def _node_with_binop(node: ast.AST, binop: ast.BinOp) -> ast.expr:
    if isinstance(node, ast.Call):
        node.args = [ast.Starred(value=binop, ctx=ast.Load())]
        return node
    elif isinstance(node, ast.List):
        # NOTE (mb 2018-06-29): Operands of the binop are always lists
        return binop
    elif isinstance(node, ast.Set):
        return ast.Call(func=ast.Name(id="set", ctx=ast.Load()), args=[binop], keywords=[])
    elif isinstance(node, ast.Tuple):
        return ast.Call(func=ast.Name(id="tuple", ctx=ast.Load()), args=[binop], keywords=[])
    else:
        raise TypeError(f"Unexpected node type {type(node)}")


def _is_stmtlist(nodelist: typ.Any) -> bool:
    return isinstance(nodelist, list) and all(isinstance(n, ast.stmt) for n in nodelist)


def _iter_walkable_fields(node: ast.AST) -> typ.Iterable[typ.Any]:
    for field_name, field_node in ast.iter_fields(node):
        if isinstance(field_node, ast.arguments):
            continue
        if isinstance(field_node, ast.expr_context):
            continue
        if isinstance(field_node, common.LeafNodeTypes):
            continue

        yield field_name, field_node


def _expand_stararg_g12n(node: ast.AST) -> ast.expr:
    """Convert fn(*x, *[1, 2], z) -> fn(*(list(x) + [1, 2, z])).

    NOTE (mb 2018-07-06): The goal here is to create an expression
      which is a list, by either creating
        1. a single list node
        2. a BinOp tree where all of the node.elts/args
            are converted to lists and concatenated.
    """

    if isinstance(node, ast.Call):
        elts = node.args
    elif isinstance(node, common.ContainerNodes):
        elts = node.elts
    else:
        raise TypeError(f"Unexpected node: {node}")

    operands: typ.List[ast.expr] = [ast.List(elts=[])]

    for elt in elts:
        tail_list = operands[-1]
        assert isinstance(tail_list, ast.List)
        tail_elts = tail_list.elts  # pylint:disable=no-member; yes it does

        if not isinstance(elt, ast.Starred):
            # NOTE (mb 2018-07-06): Simple case, just a new
            #   element for right leaf: fn(*x, *[1, 2], >z<)
            tail_elts.append(elt)
            continue

        val = elt.value
        if isinstance(val, common.ContainerNodes):
            # NOTE (mb 2018-07-06): Another simple case
            #   elements for right leaf: fn(*x,, >*[1, 2]<, z)
            tail_elts.extend(val.elts)
            continue

        # NOTE (mb 2018-07-06): Something which we can
        #   be only be sure must be an iterable, so we
        #   call list(x) and add it in the binop tree
        # elements for right leaf: fn(*>x<, *[1, 2], z)
        new_val_node = ast.Call(func=ast.Name(id="list", ctx=ast.Load()), args=[val], keywords=[])
        if len(tail_elts) == 0:
            operands[-1] = new_val_node
        else:
            operands.append(new_val_node)

        operands.append(ast.List(elts=[]))

    tail_list = operands[-1]
    assert isinstance(tail_list, ast.List)

    if len(tail_list.elts) == 0:  # pylint:disable=no-member; yes it does
        operands = operands[:-1]

    if len(operands) == 1:
        tail_list = operands[0]
        assert isinstance(tail_list, ast.List)
        return _node_with_elts(node, tail_list.elts)  # pylint:disable=no-member; yes it does

    if len(operands) > 1:
        binop = ast.BinOp(left=operands[0], op=ast.Add(), right=operands[1])
        for operand in operands[2:]:
            binop = ast.BinOp(left=binop, op=ast.Add(), right=operand)
        return _node_with_binop(node, binop)

    # NOTE (mb 2018-07-06): expand should not even have been
    #   invoked if there were no args/elts, so this signifies
    #   an error generating the operands or in detecting
    #   unpacking generalizations.
    raise RuntimeError("This should not happen")


class UnpackingGeneralizationsFixer(fb.FixerBase):

    version_info = common.VersionInfo(apply_since="2.0", apply_until="3.4")

    def expand_starstararg_g12n(self, node: ast.expr) -> ast.expr:
        chain_values: typ.List[ast.expr] = []
        chain_val   : ast.expr

        if isinstance(node, ast.Dict):
            for key, val in zip(node.keys, node.values):
                if key is None:
                    chain_val = val
                else:
                    chain_val = ast.Dict(keys=[key], values=[val])
                chain_values.append(chain_val)
        elif isinstance(node, ast.Call):
            for keyword in node.keywords:
                if keyword.arg is None:
                    chain_val = keyword.value
                else:
                    chain_val = ast.Dict(keys=[AstStr(s=keyword.arg)], values=[keyword.value])
                chain_values.append(chain_val)
        else:
            raise TypeError(f"Unexpected node type {node}")

        # collapse consecutive Dict chain values
        # [{"a": 1}, {"b": 2}] -> {"a": 1, "b": 2}
        collapsed_chain_values: typ.List[ast.expr] = []

        for chain_val in chain_values:
            # NOTE (mb 2018-06-30): We only look at the previous
            #   value for a Dict, but in principle we could look
            #   at any value. The question is, what happens when
            #   the same key is assigned to multiple times. The
            #   behaviour of unpacking generalizations is to :
            #
            #       raise TypeError(
            #           "Type object got multiple values for keyword argument '{}'"
            #       )
            #
            #   One could argue therefore, that the behaviour for
            #   the transpiled/fixed code (which doesn't raise a
            #   TypeError) is undefined and we can just collapse
            #   all ast.Dict objects into one, letting an
            #   arbitrary one of the multiple values win.

            if len(collapsed_chain_values) == 0:
                collapsed_chain_values.append(chain_val)
            else:
                prev_chain_val = collapsed_chain_values[-1]
                if isinstance(chain_val, ast.Dict) and isinstance(prev_chain_val, ast.Dict):
                    for key, val in zip(chain_val.keys, chain_val.values):
                        prev_chain_val.keys.append(key)
                        prev_chain_val.values.append(val)
                else:
                    collapsed_chain_values.append(chain_val)

        assert len(collapsed_chain_values) > 0
        if len(collapsed_chain_values) == 1:
            # NOTE (mb 2018-06-30): No need for itertools.chain if there's only
            #   a single value left after doing collapse
            collapsed_chain_value = collapsed_chain_values[0]
            if isinstance(node, ast.Dict):
                return collapsed_chain_value
            elif isinstance(node, ast.Call):
                node_func = node.func
                node_args = node.args
                if isinstance(node_func, ast.Name) and node_func.id == 'dict':
                    # value_node
                    return collapsed_chain_value
                else:
                    return ast.Call(
                        func=node_func,
                        args=node_args,
                        keywords=[ast.keyword(arg=None, value=collapsed_chain_value)],
                    )
            else:
                raise TypeError(f"Unexpected node type {node}")
        else:
            assert isinstance(node, ast.Call)
            self.required_imports.add(common.ImportDecl("itertools", None, None))
            chain_args = []
            for val in chain_values:
                items_func = ast.Attribute(value=val, attr='items', ctx=ast.Load())
                chain_args.append(ast.Call(func=items_func, args=[], keywords=[]))

            value_node = ast.Call(
                func=ast.Name(id='dict', ctx=ast.Load()),
                args=[
                    ast.Call(
                        func=ast.Attribute(
                            value=ast.Name(id='itertools', ctx=ast.Load()),
                            attr='chain',
                            ctx=ast.Load(),
                        ),
                        args=chain_args,
                        keywords=[],
                    )
                ],
                keywords=[],
            )

            node.keywords = [ast.keyword(arg=None, value=value_node)]
        return node

    def visit_expr(self, node: ast.expr) -> ast.expr:
        new_node = node
        if isinstance(node, ArgUnpackNodes) and _has_stararg_g12n(node):
            new_node = _expand_stararg_g12n(new_node)
        if isinstance(node, KwArgUnpackNodes) and _has_starstarargs_g12n(node):
            new_node = self.expand_starstararg_g12n(new_node)
        return new_node

    def walk_stmtlist(self, stmtlist: typ.List[ast.stmt]) -> typ.List[ast.stmt]:
        assert _is_stmtlist(stmtlist)

        new_stmts: typ.List[ast.stmt] = []
        for stmt in stmtlist:
            new_stmt = self.walk_stmt(stmt)
            new_stmts.append(new_stmt)
        return new_stmts

    def walk_node(self, node: ast.AST) -> ast.AST:
        if isinstance(node, common.LeafNodeTypes):
            return node

        for field_name, field_node in _iter_walkable_fields(node):
            if isinstance(field_node, ast.AST):
                new_node = self.walk_node(field_node)
                setattr(node, field_name, new_node)
            elif isinstance(field_node, list):
                new_field_node = []
                new_sub_node: ast.AST
                for sub_node in field_node:
                    if isinstance(sub_node, common.LeafNodeTypes):
                        new_sub_node = sub_node
                    elif isinstance(sub_node, ast.AST):
                        new_sub_node = self.walk_node(sub_node)
                    else:
                        new_sub_node = sub_node
                    new_field_node.append(new_sub_node)

                setattr(node, field_name, new_field_node)

        if not isinstance(node, ast.expr):
            return node

        new_expr_node = self.visit_expr(node)

        if isinstance(new_expr_node, ast.Call):
            is_single_dict_splat = (
                _is_dict_call(new_expr_node)
                and len(new_expr_node.args    ) == 0
                and len(new_expr_node.keywords) == 1
                and new_expr_node.keywords[0].arg is None
            )
            if is_single_dict_splat:
                keyword_node = new_expr_node.keywords[0]
                if _is_dict_call(keyword_node.value) or isinstance(keyword_node.value, ast.Dict):
                    return keyword_node.value

        return new_expr_node

    def walk_stmt(self, node: ast.stmt) -> ast.stmt:
        assert not _is_stmtlist(node)

        for field_name, field_node in _iter_walkable_fields(node):
            if _is_stmtlist(field_node):
                old_field_nodelist = field_node
                new_field_nodelist = self.walk_stmtlist(old_field_nodelist)
                setattr(node, field_name, new_field_nodelist)
            elif isinstance(field_node, ast.stmt):
                new_stmt = self.walk_stmt(field_node)
                setattr(node, field_name, new_stmt)
            elif isinstance(field_node, ast.AST):
                new_node = self.walk_node(field_node)
                setattr(node, field_name, new_node)
            elif isinstance(field_node, list):
                new_field_node = []
                new_sub_node: ast.AST
                for sub_node in field_node:
                    if isinstance(sub_node, common.LeafNodeTypes):
                        new_sub_node = sub_node
                    elif isinstance(sub_node, ast.AST):
                        new_sub_node = self.walk_node(sub_node)
                    else:
                        new_sub_node = sub_node
                    new_field_node.append(new_sub_node)

                setattr(node, field_name, new_field_node)
            else:
                continue

        return node

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> ast.Module:
        tree.body = self.walk_stmtlist(tree.body)
        return tree



from lib3to6 import transpile
from lib3to6.utils import clean_whitespace


def test_parse_header_simple():
    source = clean_whitespace(
        """
        # coding: ascii
        # Header line
        expr = 1 + 1
        """
    )
    header = transpile.parse_module_header(source, "2.7")
    assert header.coding == "ascii"
    assert header.text   == "# coding: ascii\n# Header line\n"


def test_parse_header_coding():
    source = clean_whitespace(
        """
    # coding: shift_jis
    # 今日は
    expr = 1 + 1
    """
    )
    header = transpile.parse_module_header(source, "2.7")
    assert header.coding == "shift_jis"
    assert header.text   == "# coding: shift_jis\n# 今日は\n"

    source_data = source.encode("shift_jis")
    header      = transpile.parse_module_header(source_data, "2.7")
    assert header.coding == "shift_jis"
    assert header.text   == "# coding: shift_jis\n# 今日は\n"



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import ast

from . import common
from . import fixer_base as fb


class BuiltinsRenameFixerBase(fb.FixerBase):

    new_name: str
    old_name: str

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> ast.Module:
        for node in ast.walk(tree):
            is_access_to_builtin = (
                isinstance(node, ast.Name)
                and isinstance(node.ctx, ast.Load)
                and node.id == self.new_name
            )

            if is_access_to_builtin:
                self.required_imports.add(common.ImportDecl("builtins", None, "__builtin__"))
                builtin_renmae_decl_str = f"""
                {self.new_name} = getattr(builtins, '{self.old_name}', {self.new_name})
                """
                self.module_declarations.add(builtin_renmae_decl_str.strip())

        return tree


class XrangeToRangeFixer(BuiltinsRenameFixerBase):

    version_info = common.VersionInfo(apply_until="2.7")

    new_name = "range"
    old_name = "xrange"


class UnicodeToStrFixer(BuiltinsRenameFixerBase):

    version_info = common.VersionInfo(apply_until="2.7")

    new_name = "str"
    old_name = "unicode"


class UnichrToChrFixer(BuiltinsRenameFixerBase):

    version_info = common.VersionInfo(apply_until="2.7")

    new_name = "chr"
    old_name = "unichr"


class RawInputToInputFixer(BuiltinsRenameFixerBase):

    version_info = common.VersionInfo(apply_until="2.7")

    new_name = "input"
    old_name = "raw_input"



import sys
from collections import namedtuple

import pytest

from lib3to6 import utils
from lib3to6 import common

CheckFixture = namedtuple("CheckFixture", ["names", "test_source", 'expected_error_msg'])


def make_fixture(names, test_source, expected_error_msg):
    test_source = utils.clean_whitespace(test_source)
    return CheckFixture(names, test_source, expected_error_msg)


FIXTURES = [
    make_fixture(
        "no_overridden_fixer_imports",
        """
        import itertools
        """,
        None,
    ),
    make_fixture(
        "no_overridden_fixer_imports",
        """
        itertools = "foo"
        """,
        "Prohibited override of import 'itertools'",
    ),
    make_fixture(
        "no_overridden_fixer_imports",
        """
        def itertools():
            pass
        """,
        "Prohibited override of import 'itertools'",
    ),
    make_fixture(
        "no_overridden_fixer_imports",
        """
        import x as itertools
        """,
        "Prohibited override of import 'itertools'",
    ),
    make_fixture(
        "no_overridden_builtins",
        """
        def map(x):
            pass
        """,
        "Prohibited override of builtin 'map'",
    ),
    make_fixture(
        "no_overridden_builtins",
        """
        class filter(x):
            pass
        """,
        "Prohibited override of builtin 'filter'",
    ),
    make_fixture(
        "no_overridden_builtins",
        """
        zip = "foobar"
        """,
        "Prohibited override of builtin 'zip'",
    ),
    make_fixture(
        "no_overridden_builtins",
        """
        repr: str = "foobar"
        """,
        "Prohibited override of builtin 'repr'",
    ),
    make_fixture(
        "no_overridden_builtins",
        """
        import foobar as iter
        """,
        "Prohibited override of builtin 'iter'",
    ),
    make_fixture(
        "no_overridden_builtins",
        """
        import reversed
        """,
        "Prohibited override of builtin 'reversed'",
    ),
    make_fixture(
        "no_overridden_builtins",
        """
        def foo(input):
            pass
        """,
        "Prohibited override of builtin 'input'",
    ),
    make_fixture(
        "no_overridden_builtins",
        """
        def foo(min=None):
            pass
        """,
        "Prohibited override of builtin 'min'",
    ),
    make_fixture(
        "no_overridden_builtins",
        """
        def foo(*, max=None):
            pass
        """,
        "Prohibited override of builtin 'max'",
    ),
    make_fixture(
        "no_overridden_builtins",
        """
        class foo:
            dict = None
        """,
        "Prohibited override of builtin 'dict'",
    ),
    make_fixture(
        "no_overridden_builtins",
        """
        class foo:
            list:  None
        """,
        "Prohibited override of builtin 'list'",
    ),
    make_fixture(
        "no_overridden_builtins",
        """
        def foo(x: int = None):
            pass
        """,
        None,
    ),
    make_fixture(
        "no_async_await",
        """
        async def foo():
            bar()
        """,
        "Prohibited use of 'async def'",
    ),
    make_fixture(
        "no_async_await",
        """
        def foo():
            await bar()
        """,
        "Prohibited use of 'await'",
    ),
    make_fixture(
        "no_yield_from",
        """
        def foo():
            yield from bar()
        """,
        "Prohibited use of 'yield from'",
    ),
    make_fixture(
        "no_open_with_encoding",
        """
        data = open(filepath, mode="rb").read()
        assert isinstance(data, bytes)
        """,
        None,
    ),
    make_fixture(
        "no_open_with_encoding",
        """
        with open(filepath, mode="wb") as fobj:
            fobj.write(b"test")
        """,
        None,
    ),
    make_fixture(
        "no_open_with_encoding",
        """
        with open(filepath, x) as fobj:
            fobj.write("test")
        """,
        "Prohibited value for argument 'mode' of builtin.open. Expected ast.Str node",
    ),
    make_fixture(
        "no_open_with_encoding",
        """
        with open(filepath, mode=x) as fobj:
            fobj.write("test")
        """,
        "Prohibited value for argument 'mode' of builtin.open. Expected ast.Str node",
    ),
    make_fixture(
        "no_open_with_encoding",
        """
        with open(filepath, mode="w") as fobj:
            fobj.write("test")
        """,
        [
            "Prohibited value 'w' for argument 'mode' of builtin.open. ",
            "Only binary modes are allowed, use io.open as an alternative.",
        ],
    ),
    make_fixture(
        "no_open_with_encoding",
        """
        with open(filepath, mode="w", encoding="utf-8") as fobj:
            fobj.write("test")
        """,
        "Prohibited keyword argument 'encoding' to builtin.open.",
    ),
    make_fixture(
        "no_star_imports",
        """
        from math import *
        """,
        "Prohibited from math import *.",
    ),
    make_fixture(
        "no_complex_named_tuple",
        """
        import typing
        class Foo(typing.NamedTuple):
            bar: int
            baz: int = 1
        """,
        "Prohibited use of default value for field 'baz' of class 'Foo'",
    ),
    make_fixture(
        "no_complex_named_tuple",
        """
        import typing
        class Foo(typing.NamedTuple):
            bar: int
            def baz(self):
                pass
        """,
        "Prohibited definition of method 'baz' for class 'Foo'",
    ),
    make_fixture(
        "no_complex_named_tuple",
        """
        import typing as typ
        class Foo(typ.NamedTuple):
            bar: int
            baz: int = 1
        """,
        "Prohibited use of default value for field 'baz' of class 'Foo'",
    ),
    make_fixture(
        "no_unusable_imports",
        """
        import lzma
        """,
        "Prohibited import 'lzma'. This module is available since Python 3.3",
    ),
    make_fixture(
        "no_unusable_imports",
        """
        from pathlib import Path
        """,
        [
            "Prohibited import 'pathlib'. This module is available since Python 3.4",
            "Use 'https://pypi.org/project/pathlib2' instead.",
        ],
    ),
    make_fixture(
        "no_unusable_imports",
        """
        import asyncio
        """,
        ["Prohibited import 'asyncio'. This module is available since Python 3.4", "No backport"],
    ),
    make_fixture(
        "no_mat_mult_op",
        "foo = bar @ baz",
        "Prohibited use of matrix multiplication",
    ),
]


@pytest.mark.parametrize("fixture", FIXTURES)
def test_checkers(fixture):
    if "--capture=no" in sys.argv:
        print()

    if isinstance(fixture.expected_error_msg, list):
        expected_error_messages = fixture.expected_error_msg
    else:
        expected_error_messages = [fixture.expected_error_msg]

    ctx = common.init_build_context(checkers=fixture.names, install_requires=set())
    try:
        utils.transpile_and_dump(ctx, fixture.test_source)
        assert fixture.expected_error_msg is None, f"Missing CheckError for '{fixture.names}'"
    except common.CheckError as result_error:
        assert fixture.expected_error_msg is not None

        for expected_error_msg in expected_error_messages:
            assert expected_error_msg in str(result_error)


def _test_unusable_imports(source, ver="2.7", req=None):
    ctx = common.init_build_context(
        checkers="no_unusable_imports",
        target_version=ver,
        install_requires=set(req.split(" ")) if req else None,
    )
    utils.transpile_and_dump(ctx, source)


def test_backport_checker_warning(caplog):
    assert len(caplog.records) == 0
    _test_unusable_imports("import typing", "2.7")
    assert len(caplog.records) == 1
    for record in caplog.records:
        assert record.levelname == 'WARNING'
        assert "Use of import 'typing'" in record.message


def test_backport_checker_nowarning(caplog):
    assert len(caplog.records) == 0
    _test_unusable_imports("import typing", "2.7", req="typing")
    assert len(caplog.records) == 0


def test_backport_checker_errors():
    # no error expected, target is new enough
    _test_unusable_imports("import asyncio", "3.4")

    try:
        _test_unusable_imports("import asyncio", "3.3")
        assert False, "expected CheckError"
    except common.CheckError as err:
        assert "Prohibited import 'asyncio'" in str(err)

    # no error expected, target new enough
    _test_unusable_imports("import pathlib", "3.4")
    # no error expected, backport used
    _test_unusable_imports("import pathlib2 as pathlib", "2.7", req="pathlib2")

    try:
        _test_unusable_imports("import pathlib", "2.7")
        assert False, "expected CheckError"
    except common.CheckError as err:
        assert "Prohibited import 'pathlib'" in str(err)

    try:
        _test_unusable_imports("import pathlib", "2.7", req="pathlib")
        assert False, "expected CheckError"
    except common.CheckError as err:
        # invalid backport
        assert "Prohibited import 'pathlib'" in str(err)

    try:
        _test_unusable_imports("import pathlib", "2.7", req="pathlib2")
        assert False, "expected CheckError"
    except common.CheckError as err:
        # use of stdlib instead of backported module
        assert "Prohibited import 'pathlib'" in str(err)

    # no error expected, target new enough
    _test_unusable_imports("import lzma", "3.3")
    # no error expected, backport used
    _test_unusable_imports("import lzma", "2.7", req="backports.lzma")

    try:
        _test_unusable_imports("import lzma", "2.7", req="lzma")
        assert False, "expected CheckError"
    except common.CheckError as err:
        # invalid backport requirement/package
        assert "Prohibited import 'lzma'" in str(err)



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import ast

from . import common
from . import fixer_base as fb


class FutureImportFixerBase(fb.FixerBase):

    future_name: str

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> ast.Module:
        self.required_imports.add(common.ImportDecl("__future__", self.future_name, None))
        return tree


class AnnotationsFutureFixer(FutureImportFixerBase):

    version_info = common.VersionInfo(apply_since="3.7", apply_until="3.99")

    future_name = "annotations"


class GeneratorStopFutureFixer(FutureImportFixerBase):

    version_info = common.VersionInfo(apply_since="3.5", apply_until="3.6")

    future_name = "generator_stop"


class UnicodeLiteralsFutureFixer(FutureImportFixerBase):

    version_info = common.VersionInfo(apply_since="2.6", apply_until="2.7")

    future_name = "unicode_literals"


class PrintFunctionFutureFixer(FutureImportFixerBase):

    version_info = common.VersionInfo(apply_since="2.6", apply_until="2.7")

    future_name = "print_function"


class WithStatementFutureFixer(FutureImportFixerBase):

    version_info = common.VersionInfo(apply_since="2.5", apply_until="2.5")

    future_name = "with_statement"


class AbsoluteImportFutureFixer(FutureImportFixerBase):

    version_info = common.VersionInfo(apply_since="2.5", apply_until="2.7")

    future_name = "absolute_import"


class DivisionFutureFixer(FutureImportFixerBase):

    version_info = common.VersionInfo(apply_since="2.2", apply_until="2.7")

    future_name = "division"


class GeneratorsFutureFixer(FutureImportFixerBase):

    version_info = common.VersionInfo(apply_since="2.2", apply_until="2.2")

    future_name = "generators"


class NestedScopesFutureFixer(FutureImportFixerBase):

    version_info = common.VersionInfo(apply_since="2.1", apply_until="2.1")

    future_name = "nested_scopes"


class RemoveUnsupportedFuturesFixer(fb.FixerBase):

    version_info = common.VersionInfo(apply_since="2.0", apply_until="3.99")

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> ast.Module:
        target_version    = ctx.cfg.target_version
        supported_futures = set()
        for cls in FutureImportFixerBase.__subclasses__():
            # pylint:disable=no-member; yes it does
            if cls.version_info.is_compatible_with(target_version):
                supported_futures.add(cls.future_name)

        nodes_to_del = []
        for i, node in enumerate(tree.body):
            is_doc_string = isinstance(node, ast.Expr) and (
                isinstance(node.value, (ast.Constant, ast.Str))
            )
            if is_doc_string:
                continue

            if not isinstance(node, ast.ImportFrom):
                break

            is_future_import = node.module == '__future__' and node.level == 0
            if not is_future_import:
                break

            new_names = [alias for alias in node.names if alias.name in supported_futures]
            if new_names == node.names:
                continue

            if not any(new_names):
                nodes_to_del.append(i)
            else:
                node.names = new_names

        for i in reversed(nodes_to_del):
            del tree.body[i : i + 1]

        return tree



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import ast
import typing as typ

from . import common
from . import fixer_base as fb


class FStringToStrFormatFixer(fb.TransformerFixerBase):

    version_info = common.VersionInfo(apply_since="2.6", apply_until="3.5")

    def _formatted_value_str(
        self, fmt_val_node: ast.FormattedValue, arg_nodes: typ.List[ast.expr]
    ) -> str:
        arg_index = len(arg_nodes)
        arg_nodes.append(fmt_val_node.value)

        format_spec_node = fmt_val_node.format_spec
        if format_spec_node is None:
            format_spec = ""
        elif not isinstance(format_spec_node, ast.JoinedStr):
            raise common.FixerError("Unexpected Node Type", format_spec_node)
        else:
            format_spec = ":" + self._joined_str_str(format_spec_node, arg_nodes)

        return "{" + str(arg_index) + format_spec + "}"

    def _joined_str_str(self, joined_str_node: ast.JoinedStr, arg_nodes: typ.List[ast.expr]) -> str:
        fmt_str = ""
        for val in joined_str_node.values:
            if isinstance(val, ast.Str):
                fmt_str += val.s
            elif isinstance(val, ast.FormattedValue):
                fmt_str += self._formatted_value_str(val, arg_nodes)
            else:
                raise common.FixerError("Unexpected Node Type", val)
        return fmt_str

    def visit_JoinedStr(self, node: ast.JoinedStr) -> ast.Call:
        arg_nodes: typ.List[ast.expr] = []

        fmt_str          = self._joined_str_str(node, arg_nodes)
        format_attr_node = ast.Attribute(value=ast.Str(s=fmt_str), attr="format", ctx=ast.Load())
        return ast.Call(func=format_attr_node, args=arg_nodes, keywords=[])



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import os
import sys
import shutil
import typing as typ
import hashlib as hl
import tempfile

import pathlib2 as pl

from . import common
from . import transpile

ENV_PATH = str(pl.Path(sys.executable).parent.parent)


PYTHON_TAG_PREFIXES = {
    'py': "Generic Python",
    'cp': "CPython",
    'ip': "IronPython",
    'pp': "PyPy",
    'jy': "Jython",
}


CACHE_DIR = pl.Path(tempfile.gettempdir()) / ".lib3to6_cache"


def eval_build_config(**kwargs) -> common.BuildConfig:
    # TODO (mb 2018-06-07): Get options from setup.cfg
    # python_tags = "py2.py3"
    # for argi, arg in enumerate(sys.argv):
    #     if "--python-tag" in arg:
    #         if "=" in arg:
    #             python_tags = arg.split("=", 1)[-1]
    #         else:
    #             python_tags = sys.argv[argi + 1]

    target_version    = kwargs.get('target_version', transpile.DEFAULT_TARGET_VERSION)
    _install_requires = kwargs.get('install_requires', None)
    cache_enabled     = kwargs.get('cache_enabled', True)
    default_mode      = kwargs.get('default_mode', 'enabled')

    install_requires: common.InstallRequires
    if _install_requires is None:
        install_requires = None
    elif isinstance(_install_requires, str):
        install_requires = set(_install_requires.split())
    elif isinstance(_install_requires, list):
        install_requires = set(_install_requires)
    else:
        raise TypeError(f"Invalid argument for install_requires: {type(_install_requires)}")

    if install_requires:
        install_requires = {req.split(";")[0] for req in install_requires}

    return common.BuildConfig(
        target_version=target_version,
        cache_enabled=cache_enabled,
        default_mode=default_mode,
        fixers="",
        checkers="",
        install_requires=install_requires,
    )


def _ignore_tmp_files(src: str, names: typ.List[str]) -> typ.List[str]:
    if isinstance(src, str):
        src_str = src
    else:
        # https://bugs.python.org/issue39390
        if isinstance(src, os.DirEntry):
            src_str = src.path
        else:
            src_str = str(src)

    if src_str.startswith("build") or src_str.startswith("./build"):
        return names
    if src_str.endswith(".egg-info"):
        return names
    if src_str.endswith("dist"):
        return names
    if src_str.endswith("__pycache__"):
        return names

    return [name for name in names if name.endswith(".pyc")]


def init_build_package_dir(local_package_dir: common.PackageDir) -> common.PackageDir:
    output_dir = pl.Path("build") / "lib3to6_out"
    output_dir.mkdir(parents=True, exist_ok=True)

    build_package_dir: common.PackageDir = {}

    for package, src_package_dir in local_package_dir.items():
        is_abs_path = pl.Path(src_package_dir) == pl.Path(src_package_dir).absolute()
        if is_abs_path:
            raise Exception(f"package_dir must use relative paths, got '{src_package_dir}'")

        build_package_subdir = output_dir / src_package_dir

        # TODO (mb 2018-08-25): As an optimization, we could
        #   restrict deletion to files that we manipulate, in
        #   other words, to *.py files.
        if build_package_subdir.exists():
            shutil.rmtree(build_package_subdir)

        shutil.copytree(src_package_dir, str(build_package_subdir), ignore=_ignore_tmp_files)
        build_package_dir[package] = str(build_package_subdir)

    return build_package_dir


def _transpile_path(cfg: common.BuildConfig, filepath: pl.Path) -> pl.Path:
    with open(filepath, mode="rb") as fobj:
        module_source_data = fobj.read()

    filehash   = hl.sha1(module_source_data).hexdigest()
    cache_path = CACHE_DIR / (filehash + ".py")

    # NOTE (mb 2020-09-01): not cache_enabled -> always update cache
    if not cfg.cache_enabled or not cache_path.exists():
        ctx = common.BuildContext(cfg, str(filepath))
        try:
            fixed_module_source_data = transpile.transpile_module_data(ctx, module_source_data)
        except common.CheckError as err:
            loc = str(filepath)
            if err.lineno >= 0:
                loc += "@" + str(err.lineno)

            err.args = (loc + " - " + err.args[0],) + err.args[1:]
            raise

        with open(cache_path, mode="wb") as fobj:
            fobj.write(fixed_module_source_data)

    return cache_path


def build_package(cfg: common.BuildConfig, package: str, build_dir: str) -> None:
    # pylint:disable=unused-argument ; `package` is part of the public api now
    for root, _dirs, files in os.walk(build_dir):
        for filename in files:
            filepath = pl.Path(root) / filename
            if filepath.suffix != ".py":
                continue

            transpiled_path = _transpile_path(cfg, filepath)
            # overwrite original with transpiled
            shutil.copy(transpiled_path, filepath)


def build_packages(cfg: common.BuildConfig, build_package_dir: common.PackageDir) -> None:
    CACHE_DIR.mkdir(exist_ok=True)

    for package, build_dir in build_package_dir.items():
        build_package(cfg, package, build_dir)


def fix(
    package_dir     : common.PackageDir = None,
    target_version  : str = transpile.DEFAULT_TARGET_VERSION,
    install_requires: typ.List[str] = None,
    default_mode    : str = 'enabled',
) -> common.PackageDir:
    if package_dir is None:
        package_dir = {"": "."}

    build_package_dir = init_build_package_dir(package_dir)
    build_cfg         = eval_build_config(
        target_version=target_version,
        install_requires=install_requires,
        default_mode=default_mode,
    )
    build_packages(build_cfg, build_package_dir)
    return build_package_dir



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import ast

from . import common
from . import fixer_base as fb


def _try_fallback(node: ast.stmt, fallback_node: ast.stmt) -> ast.Try:
    return ast.Try(
        body=[node],
        handlers=[
            ast.ExceptHandler(
                type=ast.Name(id="ImportError", ctx=ast.Load()), name=None, body=[fallback_node]
            )
        ],
        orelse=[],
        finalbody=[],
    )


class ModuleImportFallbackFixerBase(fb.TransformerFixerBase):

    new_name: str
    old_name: str

    def visit_Import(self, node: ast.Import) -> ast.stmt:
        if len(node.names) != 1:
            return node

        alias = node.names[0]
        if alias.name != self.new_name:
            return node

        if alias.asname:
            asname = alias.asname
        elif "." in self.new_name:
            asname = self.new_name.replace(".", "_")
            msg    = (
                f"Prohibited use of 'import {self.new_name}', "
                f"use 'import {self.new_name} as {asname}' instead."
            )
            raise common.CheckError(msg, node)
        else:
            asname = self.new_name

        return _try_fallback(node, ast.Import(names=[ast.alias(name=self.old_name, asname=asname)]))

    def visit_ImportFrom(self, node: ast.ImportFrom) -> ast.stmt:
        if node.module != self.new_name:
            return node

        return _try_fallback(
            node, ast.ImportFrom(module=self.old_name, names=node.names, level=node.level)
        )


class ConfigParserImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "configparser"
    old_name     = "ConfigParser"


class SocketServerImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "socketserver"
    old_name     = "SocketServer"


class BuiltinsImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "builtins"
    old_name     = "__builtin__"


class QueueImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "queue"
    old_name     = "Queue"


class CopyRegImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "copyreg"
    old_name     = "copy_reg"


class WinRegImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "winreg"
    old_name     = "_winreg"


class ReprLibImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "reprlib"
    old_name     = "repr"


class ThreadImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "_thread"
    old_name     = "thread"


class DummyThreadImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "_dummy_thread"
    old_name     = "dummy_thread"


# NOTE (mb 2018-09-01): Up to here are the simple cases. Below
#   here, the fixes only work when using ast.FromImport, or when
#   using asname. For everything else, we raise a CheckError. The
#   only other option would be to scan whole tree and rewrite
#   reverences


class HttpCookiejarImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "http.cookiejar"
    old_name     = "cookielib"


class UrllibParseImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "urllib.parse"
    old_name     = "urlparse"


class UrllibRequestImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "urllib.request"
    old_name     = "urllib2"


class UrllibErrorImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "urllib.error"
    old_name     = "urllib2"


class UrllibRobotParserImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "urllib.robotparser"
    old_name     = "robotparser"


class XMLRPCClientImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "xmlrpc.client"
    old_name     = "xmlrpclib"


class XmlrpcServerImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "xmlrpc.server"
    old_name     = "SimpleXMLRPCServer"


class HtmlParserImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "html.parser"
    old_name     = "HTMLParser"


class HttpClientImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "http.client"
    old_name     = "httplib"


class HttpCookiesImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "http.cookies"
    old_name     = "Cookie"


class PickleImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "pickle"
    old_name     = "cPickle"


class DbmGnuImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "dbm.gnu"
    old_name     = "gdbm"


class EmailMimeBaseImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "email.mime.base"
    old_name     = "email.MIMEBase"


class EmailMimeImageImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "email.mime.image"
    old_name     = "email.MIMEImage"


class EmailMimeMultipartImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "email.mime.multipart"
    old_name     = "email.MIMEMultipart"


class EmailMimeNonmultipartImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "email.mime.nonmultipart"
    old_name     = "email.MIMENonMultipart"


class EmailMimeTextImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "email.mime.text"
    old_name     = "email.MIMEText"


class TkinterImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "tkinter"
    old_name     = "Tkinter"


class TkinterDialogImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "tkinter.dialog"
    old_name     = "Dialog"


class TkinterScrolledTextImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "tkinter.scrolledtext"
    old_name     = "ScrolledText"


class TkinterTixImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "tkinter.tix"
    old_name     = "Tix"


class TkinterTtkImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "tkinter.ttk"
    old_name     = "ttk"


class TkinterConstantsImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "tkinter.constants"
    old_name     = "Tkconstants"


class TkinterDndImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "tkinter.dnd"
    old_name     = "Tkdnd"


class TkinterColorchooserImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "tkinter.colorchooser"
    old_name     = "tkColorChooser"


class TkinterCommonDialogImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "tkinter.commondialog"
    old_name     = "tkCommonDialog"


class TkinterFontImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "tkinter.font"
    old_name     = "tkFont"


class TkinterMessageboxImportFallbackFixer(ModuleImportFallbackFixerBase):

    version_info = common.VersionInfo(apply_since="2.3", apply_until="2.7")
    new_name     = "tkinter.messagebox"
    old_name     = "tkMessageBox"


# TODO (mb 2018-09-02): Only ImportFrom is permitted for
#   certain imports, so that we can determine which old
#   module to import from
#
#     new_name = "http.server"
#     old_name = "BaseHTTPServer"
#     old_name = "CGIHTTPServer"
#     old_name = "SimpleHTTPServer"
#
#     new_name = "tkinter.simpledialog"
#     old_name = "tkSimpleDialog"
#     old_name = "SimpleDialog"
#
#     new_name = "tkinter.filedialog"
#     old_name = "tkFileDialog"
#     old_name = "FileDialog"



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import ast
import sys
import typing as typ

from . import utils
from . import common
from . import fixer_base as fb
from .fixers_future import DivisionFutureFixer
from .fixers_future import GeneratorsFutureFixer
from .fixers_future import AnnotationsFutureFixer
from .fixers_future import NestedScopesFutureFixer
from .fixers_future import GeneratorStopFutureFixer
from .fixers_future import PrintFunctionFutureFixer
from .fixers_future import WithStatementFutureFixer
from .fixers_future import AbsoluteImportFutureFixer
from .fixers_future import UnicodeLiteralsFutureFixer
from .fixers_future import RemoveUnsupportedFuturesFixer
from .fixers_builtin_rename import UnichrToChrFixer
from .fixers_builtin_rename import UnicodeToStrFixer
from .fixers_builtin_rename import XrangeToRangeFixer
from .fixers_builtin_rename import RawInputToInputFixer
from .fixers_import_fallback import QueueImportFallbackFixer
from .fixers_import_fallback import DbmGnuImportFallbackFixer
from .fixers_import_fallback import PickleImportFallbackFixer
from .fixers_import_fallback import ThreadImportFallbackFixer
from .fixers_import_fallback import WinRegImportFallbackFixer
from .fixers_import_fallback import CopyRegImportFallbackFixer
from .fixers_import_fallback import ReprLibImportFallbackFixer
from .fixers_import_fallback import TkinterImportFallbackFixer
from .fixers_import_fallback import BuiltinsImportFallbackFixer
from .fixers_import_fallback import HtmlParserImportFallbackFixer
from .fixers_import_fallback import HttpClientImportFallbackFixer
from .fixers_import_fallback import TkinterDndImportFallbackFixer
from .fixers_import_fallback import TkinterTixImportFallbackFixer
from .fixers_import_fallback import TkinterTtkImportFallbackFixer
from .fixers_import_fallback import DummyThreadImportFallbackFixer
from .fixers_import_fallback import HttpCookiesImportFallbackFixer
from .fixers_import_fallback import TkinterFontImportFallbackFixer
from .fixers_import_fallback import UrllibErrorImportFallbackFixer
from .fixers_import_fallback import UrllibParseImportFallbackFixer
from .fixers_import_fallback import ConfigParserImportFallbackFixer
from .fixers_import_fallback import SocketServerImportFallbackFixer
from .fixers_import_fallback import XMLRPCClientImportFallbackFixer
from .fixers_import_fallback import XmlrpcServerImportFallbackFixer
from .fixers_import_fallback import EmailMimeBaseImportFallbackFixer
from .fixers_import_fallback import EmailMimeTextImportFallbackFixer
from .fixers_import_fallback import HttpCookiejarImportFallbackFixer
from .fixers_import_fallback import TkinterDialogImportFallbackFixer
from .fixers_import_fallback import UrllibRequestImportFallbackFixer
from .fixers_import_fallback import EmailMimeImageImportFallbackFixer
from .fixers_import_fallback import TkinterConstantsImportFallbackFixer
from .fixers_import_fallback import TkinterMessageboxImportFallbackFixer
from .fixers_import_fallback import UrllibRobotParserImportFallbackFixer
from .fixers_import_fallback import EmailMimeMultipartImportFallbackFixer
from .fixers_import_fallback import TkinterColorchooserImportFallbackFixer
from .fixers_import_fallback import TkinterCommonDialogImportFallbackFixer
from .fixers_import_fallback import TkinterScrolledTextImportFallbackFixer
from .fixers_import_fallback import EmailMimeNonmultipartImportFallbackFixer
from .fixers_unpacking_generalization import UnpackingGeneralizationsFixer

AstStr = getattr(ast, 'Str', ast.Constant)


def is_const_node(node: ast.AST) -> bool:
    return node is None or any(isinstance(node, cntype) for cntype in common.ConstantNodeTypes)


Elt  = typ.Union[ast.Name, ast.Constant, ast.Subscript]
Elts = typ.List[Elt]


AnnoNode = typ.Union[ast.arg, ast.AnnAssign, ast.FunctionDef]


class _FRAFContext:

    local_classes: typ.Set[str]
    known_classes: typ.Set[str]

    def __init__(self, local_classes: typ.Set[str]) -> None:
        self.local_classes = local_classes
        self.known_classes = set()

    def is_forward_ref(self, name: str) -> bool:
        return name in self.local_classes and name not in self.known_classes

    def update_index_elts(self, elts: Elts) -> None:
        # NOTE (mb 2020-07-19): We modify elts during iteration
        #   pylint:disable=consider-using-enumerate
        for i in range(len(elts)):
            elt = elts[i]
            if is_const_node(elt) or isinstance(elt, ast.Attribute):
                continue

            if isinstance(elt, ast.Name):
                if self.is_forward_ref(elt.id):
                    elts[i] = ast.Constant(elt.id)
            elif isinstance(elt, ast.Subscript):
                idx = elt.slice
                assert isinstance(idx, ast.Index)
                self.update_index(idx)
            else:
                msg = f"Error fixing index element with forward ref of type {type(elt)}"
                raise NotImplementedError(msg)

    def update_index(self, idx: ast.Index) -> None:
        val = idx.value
        if is_const_node(val) or isinstance(val, ast.Attribute):
            return

        if isinstance(val, ast.Name):
            if self.is_forward_ref(val.id):
                idx.value = AstStr(val.id)
        elif isinstance(val, ast.Subscript):
            sub_idx = val.slice
            assert isinstance(sub_idx, ast.Index)
            self.update_index(sub_idx)
        elif isinstance(val, ast.Tuple):
            elts = typ.cast(Elts, val.elts)
            self.update_index_elts(elts)
        else:
            msg = f"Error fixing index with forward ref of type {type(val)}"
            raise NotImplementedError(msg)

    def update_annotation_refs(self, node: AnnoNode, attrname: str) -> None:
        anno = getattr(node, attrname)
        if is_const_node(anno) or isinstance(anno, ast.Attribute):
            return

        if isinstance(anno, ast.Name):
            if self.is_forward_ref(anno.id):
                setattr(node, attrname, AstStr(anno.id))
        elif isinstance(anno, ast.Subscript):
            idx = anno.slice
            assert isinstance(idx, ast.Index)
            self.update_index(idx)
        else:
            msg = f"Error fixing annotation of forward ref with type {type(anno)}"
            raise NotImplementedError(msg)

    def remove_forward_references(self, node: ast.AST) -> None:
        for sub_node in ast.iter_child_nodes(node):
            if isinstance(sub_node, ast.FunctionDef):
                self.update_annotation_refs(sub_node, 'returns')

                for arg in sub_node.args.args:
                    self.update_annotation_refs(arg, 'annotation')
                for arg in sub_node.args.kwonlyargs:
                    self.update_annotation_refs(arg, 'annotation')

                kwarg = sub_node.args.kwarg
                if kwarg:
                    self.update_annotation_refs(kwarg, 'annotation')
                vararg = sub_node.args.vararg
                if vararg:
                    self.update_annotation_refs(vararg, 'annotation')
            elif isinstance(sub_node, ast.AnnAssign):
                self.update_annotation_refs(sub_node, 'annotation')

            if hasattr(sub_node, 'body'):
                self.remove_forward_references(sub_node)

            if isinstance(sub_node, ast.ClassDef):
                self.known_classes.add(sub_node.name)


class ForwardReferenceAnnotationsFixer(fb.FixerBase):

    version_info = common.VersionInfo(apply_since="3.0", apply_until="3.6")

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> ast.Module:
        local_classes: typ.Set[str] = set()
        for node in ast.walk(tree):
            if isinstance(node, ast.ClassDef):
                local_classes.add(node.name)

        _FRAFContext(local_classes).remove_forward_references(tree)
        return tree


class RemoveFunctionDefAnnotationsFixer(fb.FixerBase):

    version_info = common.VersionInfo(apply_since="1.0", apply_until="2.7")

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> ast.Module:
        for node in ast.walk(tree):
            if isinstance(node, ast.FunctionDef):
                node.returns = None
                for arg in node.args.args:
                    arg.annotation = None
                for arg in node.args.kwonlyargs:
                    arg.annotation = None

                if node.args.kwarg:
                    node.args.kwarg.annotation = None
                if node.args.vararg:
                    node.args.vararg.annotation = None

        return tree


class RemoveAnnAssignFixer(fb.TransformerFixerBase):

    version_info = common.VersionInfo(apply_since="1.0", apply_until="3.5")

    @staticmethod
    def visit_AnnAssign(node: ast.AnnAssign) -> ast.Assign:
        tgt_node = node.target
        if not isinstance(tgt_node, (ast.Name, ast.Attribute)):
            raise common.FixerError("Unexpected Node type", tgt_node)

        value: ast.expr
        if node.value is None:
            value = ast.NameConstant(value=None)
        else:
            value = node.value
        return ast.Assign(targets=[tgt_node], value=value)


class ShortToLongFormSuperFixer(fb.TransformerFixerBase):

    version_info = common.VersionInfo(apply_since="2.2", apply_until="2.7")

    @staticmethod
    def visit_ClassDef(node: ast.ClassDef) -> ast.ClassDef:
        for maybe_method in ast.walk(node):
            if not isinstance(maybe_method, ast.FunctionDef):
                continue

            method     : ast.FunctionDef = maybe_method
            method_args: ast.arguments   = method.args
            if len(method_args.args) == 0:
                continue

            self_arg: ast.arg = method_args.args[0]

            for maybe_super_call in ast.walk(method):
                if not isinstance(maybe_super_call, ast.Call):
                    continue

                func_node = maybe_super_call.func
                if not (isinstance(func_node, ast.Name) and func_node.id == "super"):
                    continue

                super_call = maybe_super_call
                if len(super_call.args) > 0:
                    continue

                super_call.args = [
                    ast.Name(id=node.name   , ctx=ast.Load()),
                    ast.Name(id=self_arg.arg, ctx=ast.Load()),
                ]
        return node


class InlineKWOnlyArgsFixer(fb.TransformerFixerBase):

    version_info = common.VersionInfo(apply_since="1.0", apply_until="2.99")

    @staticmethod
    def visit_FunctionDef(node: ast.FunctionDef) -> ast.FunctionDef:
        if not node.args.kwonlyargs:
            return node

        if node.args.kwarg:
            kw_name = node.args.kwarg.arg
        else:
            kw_name         = "kwargs"
            node.args.kwarg = ast.arg(arg=kw_name, annotation=None)

        kwonlyargs  = reversed(node.args.kwonlyargs)
        kw_defaults = reversed(node.args.kw_defaults)
        for arg, default in zip(kwonlyargs, kw_defaults):
            arg_name = arg.arg
            node_value: ast.expr

            # NOTE (mb 2018-06-03): Only use defaults for kwargs
            #   if they are literals. Everything else would
            #   change the semantics too much and so we should
            #   raise an error.
            if default is None:
                node_value = ast.Subscript(
                    value=ast.Name(id=kw_name, ctx=ast.Load()),
                    slice=ast.Index(value=AstStr(s=arg_name)),
                    ctx=ast.Load(),
                )
            elif not isinstance(default, common.ConstantNodeTypes):
                msg = (
                    f"Keyword only arguments must be immutable. " f"Found: {default} for {arg_name}"
                )
                raise common.FixerError(msg, node)
            else:
                node_value = ast.Call(
                    func=ast.Attribute(
                        value=ast.Name(id=kw_name, ctx=ast.Load()), attr="get", ctx=ast.Load()
                    ),
                    args=[AstStr(s=arg_name), default],
                    keywords=[],
                )

            new_node = ast.Assign(
                targets=[ast.Name(id=arg_name, ctx=ast.Store())], value=node_value
            )

            node.body.insert(0, new_node)

        node.args.kwonlyargs = []

        return node


class NewStyleClassesFixer(fb.TransformerFixerBase):

    version_info = common.VersionInfo(apply_since="2.0", apply_until="2.7")

    def visit_ClassDef(self, node: ast.ClassDef) -> ast.ClassDef:
        self.generic_visit(node)
        if len(node.bases) == 0:
            node.bases.append(ast.Name(id="object", ctx=ast.Load()))
        return node


class ItertoolsBuiltinsFixer(fb.TransformerFixerBase):

    version_info = common.VersionInfo(
        apply_since="2.3",  # introduction of the itertools module
        apply_until="2.7",
        works_until="3.99",
    )

    # WARNING (mb 2018-06-09): This fix is very broad, and should
    #   only be used in combination with a sanity check that the
    #   builtin names are not being overridden.

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> ast.Module:
        new_tree = self.visit(tree)
        return typ.cast(ast.Module, new_tree)

    def visit_Name(self, node: ast.Name) -> typ.Union[ast.Name, ast.Attribute]:
        if isinstance(node.ctx, ast.Load) and node.id in ("map", "zip", "filter"):
            self.required_imports.add(common.ImportDecl("itertools", None, None))
            global_decl = f"{node.id} = getattr(itertools, 'i{node.id}', {node.id})"
            self.module_declarations.add(global_decl)

        return node


class NamedTupleClassToAssignFixer(fb.TransformerFixerBase):

    version_info = common.VersionInfo(apply_since="2.6", apply_until="3.4")

    _typing_module_name   : typ.Optional[str]
    _namedtuple_class_name: typ.Optional[str]

    def __init__(self) -> None:
        self._typing_module_name    = None
        self._namedtuple_class_name = None
        super().__init__()

    def visit_ImportFrom(self, node: ast.ImportFrom) -> ast.ImportFrom:
        if node.module == 'typing':
            for alias in node.names:
                if alias.name == 'NamedTuple':
                    if alias.asname is None:
                        self._namedtuple_class_name = alias.name
                    else:
                        self._namedtuple_class_name = alias.asname

        return node

    def visit_Import(self, node: ast.Import) -> ast.Import:
        for alias in node.names:
            if alias.name == 'typing':
                if alias.asname is None:
                    self._typing_module_name = alias.name
                else:
                    self._typing_module_name = alias.asname
        return node

    def visit_ClassDef(self, node: ast.ClassDef) -> typ.Union[ast.ClassDef, ast.Assign]:
        self.generic_visit(node)
        if len(node.bases) == 0:
            return node

        if not (self._typing_module_name or self._namedtuple_class_name):
            # no import of typing.NamedTuple -> class cannot have it as one its bases
            return node

        has_namedtuple_base = utils.has_base_class(
            node, self._typing_module_name, self._namedtuple_class_name or 'NamedTuple'
        )
        if not has_namedtuple_base:
            return node

        func: typ.Union[ast.Attribute, ast.Name]

        if self._typing_module_name:
            func = ast.Attribute(
                value=ast.Name(id=self._typing_module_name, ctx=ast.Load()),
                attr="NamedTuple",
                ctx=ast.Load(),
            )
        elif self._namedtuple_class_name:
            func = ast.Name(id=self._namedtuple_class_name, ctx=ast.Load())
        else:
            raise RuntimeError("")

        elts: typ.List[ast.Tuple] = []

        for assign in node.body:
            if not isinstance(assign, ast.AnnAssign):
                continue
            tgt = assign.target
            if not isinstance(tgt, ast.Name):
                continue

            elts.append(ast.Tuple(elts=[AstStr(s=tgt.id), assign.annotation], ctx=ast.Load()))

        return ast.Assign(
            targets=[ast.Name(id=node.name, ctx=ast.Store())],
            value=ast.Call(
                func=func,
                args=[AstStr(s=node.name), ast.List(elts=elts, ctx=ast.Load())],
                keywords=[],
            ),
        )


if sys.version_info >= (3, 6):
    from .fixers_fstring import FStringToStrFormatFixer
else:
    FStringToStrFormatFixer = None


if sys.version_info >= (3, 8):
    from .fixers_namedexpr import NamedExprFixer
else:
    NamedExprFixer = None


__all__ = [
    "AnnotationsFutureFixer",
    "GeneratorStopFutureFixer",
    "UnicodeLiteralsFutureFixer",
    "RemoveUnsupportedFuturesFixer",
    "PrintFunctionFutureFixer",
    "WithStatementFutureFixer",
    "AbsoluteImportFutureFixer",
    "DivisionFutureFixer",
    "GeneratorsFutureFixer",
    "NestedScopesFutureFixer",
    "ConfigParserImportFallbackFixer",
    "SocketServerImportFallbackFixer",
    "BuiltinsImportFallbackFixer",
    "QueueImportFallbackFixer",
    "CopyRegImportFallbackFixer",
    "WinRegImportFallbackFixer",
    "ReprLibImportFallbackFixer",
    "ThreadImportFallbackFixer",
    "DummyThreadImportFallbackFixer",
    "HttpCookiejarImportFallbackFixer",
    "UrllibParseImportFallbackFixer",
    "UrllibRequestImportFallbackFixer",
    "UrllibErrorImportFallbackFixer",
    "UrllibRobotParserImportFallbackFixer",
    "XMLRPCClientImportFallbackFixer",
    "XmlrpcServerImportFallbackFixer",
    "HtmlParserImportFallbackFixer",
    "HttpClientImportFallbackFixer",
    "HttpCookiesImportFallbackFixer",
    "PickleImportFallbackFixer",
    "DbmGnuImportFallbackFixer",
    "EmailMimeBaseImportFallbackFixer",
    "EmailMimeImageImportFallbackFixer",
    "EmailMimeMultipartImportFallbackFixer",
    "EmailMimeNonmultipartImportFallbackFixer",
    "EmailMimeTextImportFallbackFixer",
    "TkinterImportFallbackFixer",
    "TkinterDialogImportFallbackFixer",
    "TkinterScrolledTextImportFallbackFixer",
    "TkinterTixImportFallbackFixer",
    "TkinterTtkImportFallbackFixer",
    "TkinterConstantsImportFallbackFixer",
    "TkinterDndImportFallbackFixer",
    "TkinterColorchooserImportFallbackFixer",
    "TkinterCommonDialogImportFallbackFixer",
    "TkinterFontImportFallbackFixer",
    "TkinterMessageboxImportFallbackFixer",
    "XrangeToRangeFixer",
    "UnicodeToStrFixer",
    "UnichrToChrFixer",
    "RawInputToInputFixer",
    "RemoveFunctionDefAnnotationsFixer",
    "ForwardReferenceAnnotationsFixer",
    "RemoveAnnAssignFixer",
    "ShortToLongFormSuperFixer",
    "InlineKWOnlyArgsFixer",
    "NewStyleClassesFixer",
    "ItertoolsBuiltinsFixer",
    "UnpackingGeneralizationsFixer",
    "NamedTupleClassToAssignFixer",
    "FStringToStrFormatFixer",
    "NamedExprFixer",
    "UnpackingGeneralizationsFixer",
]

# class GeneratorReturnToStopIterationExceptionFixer(fb.FixerBase):
#
#     version_info = common.VersionInfo(
#         apply_since="2.0",
#         apply_until="3.3",
#     )
#
#     def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> ast.Module:
#         return tree
#
#     def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
#         # NOTE (mb 2018-06-15): What about a generator nested in a function definition?
#         is_generator = any(
#             isinstance(sub_node, (ast.Yield, ast.YieldFrom))
#             for sub_node in ast.walk(node)
#         )
#         if not is_generator:
#             return node
#
#         for sub_node in ast.walk(node):
#             pass


# YIELD_FROM_EQUIVALENT = """
# _i = iter(EXPR)
# try:
#     _y = next(_i)
# except StopIteration as _e:
#     _r = _e.value
# else:
#     while 1:
#         try:
#             _s = yield _y
#         except GeneratorExit as _e:
#             try:
#                 _m = _i.close
#             except AttributeError:
#                 pass
#             else:
#                 _m()
#             raise _e
#         except BaseException as _e:
#             _x = sys.exc_info()
#             try:
#                 _m = _i.throw
#             except AttributeError:
#                 raise _e
#             else:
#                 try:
#                     _y = _m(*_x)
#                 except StopIteration as _e:
#                     _r = _e.value
#                     break
#         else:
#             try:
#                 if _s is None:
#                     _y = next(_i)
#                 else:
#                     _y = _i.send(_s)
#             except StopIteration as _e:
#                 _r = _e.value
#                 break
# RESULT = _r
# """in_len_body


# class YieldFromFixer(fb.FixerBase):
# # see https://www.python.org/dev/peps/pep-0380/
# NOTE (mb 2018-06-14): We should definetly do the most simple case
#   but maybe we can also detect the more complex cases involving
#   send and return values and at least throw an error

# class MetaclassFixer(fb.TransformerFixerBase):
#
#     version_info = common.VersionInfo(
#         apply_since="2.0",
#         apply_until="2.7",
#     )
#
#     def visit_ClassDef(self, node: ast.ClassDef) -> ast.ClassDef:
#         #  class Foo(metaclass=X): => class Foo(object):\n  __metaclass__ = X


# class MatMulFixer(fb.TransformerFixerBase):
#
#     version_info = common.VersionInfo(
#         apply_since="2.0",
#         apply_until="3.5",
#     )
#
#     def visit_Binop(self, node: ast.BinOp) -> ast.Call:
#         # replace a @ b with a.__matmul__(b)


# NOTE (mb 2018-06-24): I'm not gonna do it, but feel free to
#   implement it if you feel like it.
#
# class DecoratorFixer(fb.FixerBase):
#     """Replaces use of @decorators with function calls
#
#     > @mydec1()
#     > @mydec2
#     > def myfn():
#     >     pass
#     < def myfn():
#     <     pass
#     < myfn = mydec2(myfn)
#     < myfn = mydec1()(myfn)
#     """
#
#     version_info = common.VersionInfo(
#         apply_since="2.0",
#         apply_until="2.4",
#     )
#

# NOTE (mb 2018-06-24): I'm not gonna do it, but feel free to
#   implement it if you feel like it.
#
# class WithStatementToTryExceptFixer(fb.FixerBase):
#     """
#     > with expression as name:
#     >     name
#
#     < import sys
#     < __had_exception = False
#     < __manager = expression
#     < try:
#     <     name = manager.__enter__()
#     < except:
#     <     __had_exception = True
#     <     ex_type, ex_value, traceback = sys.exc_info()
#     <     __manager.__exit__(ex_type, ex_value, traceback)
#     < finally:
#     <     if not __had_exception:
#     <         __manager.__exit__(None, None, None)
#     """
#
#     version_info = common.VersionInfo(
#         apply_since="2.0",
#         apply_until="2.4",
#     )
#

# NOTE (mb 2018-06-25): I'm not gonna do it, but feel free to
#   implement it if you feel like it.
#
# class ImplicitFormatIndexesFixer(fb.FixerBase):
#     """Replaces use of @decorators with function calls
#
#     > "first: {} second: {:>9}".format(0, 1)
#     < "first: {0} second: {1:>9}".format(0, 1)
#     """
#
#     version_info = common.VersionInfo(
#         apply_since="2.6",
#         apply_until="2.6",
#     )
#



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import re
import ast
import sys
import typing as typ

import astor

from . import utils
from . import common
from . import fixers
from . import checkers
from . import fixer_base as fb
from . import checker_base as cb

DEFAULT_SOURCE_ENCODING_DECLARATION = "# -*- coding: {} -*-"

DEFAULT_SOURCE_ENCODING = "utf-8"

DEFAULT_TARGET_VERSION = "2.7"

# https://www.python.org/dev/peps/pep-0263/
SOURCE_ENCODING_PATTERN = r"""
    ^
    [ \t\v]*
    \#.*?coding[:=][ \t]*
    (?P<coding>[-_.a-zA-Z0-9]+)
"""

SOURCE_ENCODING_RE = re.compile(SOURCE_ENCODING_PATTERN, re.VERBOSE)


MODE_MARKER_PATTERN = r"#\s*lib3to6:\s*(?P<mode>disabled|enabled)"

MODE_MARKER_RE = re.compile(MODE_MARKER_PATTERN, flags=re.MULTILINE)


class ModuleHeader(typ.NamedTuple):

    coding: str
    text  : str


def _parse_header_line(line_data: typ.Union[bytes, str], coding: str) -> str:
    if isinstance(line_data, bytes):
        return line_data.decode(coding)
    if isinstance(line_data, str):
        return line_data

    # unreachable
    bad_type = type(line_data)
    errmsg   = f"Invalid type: line_data must be str/bytes but was '{bad_type}'"
    raise TypeError(errmsg)


def parse_module_header(module_source: typ.Union[bytes, str], target_version: str) -> ModuleHeader:
    shebang = False
    coding  = None
    line: str

    header_lines: typ.List[str] = []

    for i, line_data in enumerate(module_source.splitlines()):
        assert isinstance(line_data, (bytes, str))
        line = _parse_header_line(line_data, coding or DEFAULT_SOURCE_ENCODING)

        if i < 2:
            if i == 0 and line.startswith("#!") and "python" in line:
                shebang = True
            else:
                match = SOURCE_ENCODING_RE.match(line)
                if match:
                    coding = match.group("coding").strip()

        if line.rstrip() and not line.rstrip().startswith("#"):
            break

        header_lines.append(line)

    if coding is None:
        coding = DEFAULT_SOURCE_ENCODING
        if target_version < "3.0":
            coding_decl = DEFAULT_SOURCE_ENCODING_DECLARATION.format(coding)
            if shebang:
                header_lines.insert(1, coding_decl)
            else:
                header_lines.insert(0, coding_decl)

    header_text = "\n".join(header_lines) + "\n"
    return ModuleHeader(coding, header_text)


CheckerType = typ.Type[cb.CheckerBase]

FixerType = typ.Type[fb.FixerBase]

CheckerOrFixer = typ.Union[CheckerType, FixerType]


def normalize_name(name: str) -> str:
    name = name.strip().lower().replace("_", "").replace("-", "")
    if name.endswith("fixer"):
        name = name[: -len("fixer")]
    if name.endswith("checker"):
        name = name[: -len("checker")]
    return name


def get_available_classes(module: object, clazz: CheckerOrFixer) -> typ.Dict[str, CheckerOrFixer]:

    assert isinstance(clazz, type)
    clazz_name = clazz.__name__
    assert clazz_name.endswith("Base")

    maybe_classes = {
        name: getattr(module, name) for name in dir(module) if not name.endswith(clazz_name)
    }

    return {
        normalize_name(attr_name): attr
        for attr_name, attr in maybe_classes.items()
        if isinstance(attr, type) and issubclass(attr, clazz)
    }


FuzzyNames = typ.Union[str, typ.List[str]]


def get_selected_names(names: FuzzyNames, available_names: typ.Set[str]) -> typ.List[str]:
    if isinstance(names, str):
        names_list = names.split(",")
    else:
        names_list = names

    selected_names = [normalize_name(name) for name in names_list if name.strip()]

    if selected_names:
        for name in selected_names:
            assert name in available_names
    else:
        # Nothing explicitly selected -> all selected
        selected_names = sorted(available_names)

    assert len(selected_names) > 0

    return selected_names


def iter_fuzzy_selected_checkers(names: FuzzyNames) -> typ.Iterable[cb.CheckerBase]:
    available_classes = get_available_classes(checkers, cb.CheckerBase)
    selected_names    = get_selected_names(names, set(available_classes))
    for name in selected_names:
        checker_type = typ.cast(CheckerType, available_classes[name])
        yield checker_type()


def iter_fuzzy_selected_fixers(names: FuzzyNames) -> typ.Iterable[fb.FixerBase]:
    available_classes = get_available_classes(fixers, fb.FixerBase)
    selected_names    = get_selected_names(names, set(available_classes))
    for name in selected_names:
        fixer_type = typ.cast(FixerType, available_classes[name])
        yield fixer_type()


def find_import_decls(node: ast.AST) -> typ.Iterable[common.ImportDecl]:
    # NOTE (mb 2020-07-18): returns as fences are fine
    # pylint:disable=too-many-return-statements
    # NOTE (mb 2020-07-18): despite the brnaches, the code is quite linear
    # pylint:disable=too-many-branches
    if not isinstance(node, (ast.Try, ast.Import, ast.ImportFrom)):
        return

    if isinstance(node, ast.Try):
        if not (len(node.body) == 1 and len(node.handlers) == 1):
            return

        except_handler = node.handlers[0]

        is_import_error_handler = (
            isinstance(except_handler.type, ast.Name)
            and except_handler.type.id == 'ImportError'
            and len(except_handler.body) == 1
        )
        if not is_import_error_handler:
            return

        maybe_import = node.body[0]
        if not isinstance(maybe_import, ast.Import):
            return

        default_import = maybe_import

        maybe_fallback_import = except_handler.body[0]
        if not isinstance(maybe_fallback_import, ast.Import):
            return

        fallback_import = maybe_fallback_import

        if len(default_import.names) == 1 and len(fallback_import.names) == 1:
            default_import_alias  = default_import.names[0]
            fallback_import_alias = fallback_import.names[0]
            yield common.ImportDecl(
                default_import_alias.name, default_import_alias.asname, fallback_import_alias.name
            )

    elif isinstance(node, ast.Import):
        if len(node.names) != 1 and any(alias.asname for alias in node.names):
            # we never use multi name imports or asname, so this is user code
            return

        alias = node.names[0]
        yield common.ImportDecl(alias.name, None, None)
    elif isinstance(node, ast.ImportFrom):
        if any(alias.asname for alias in node.names):
            # we never use multi name imports or asname, so this is user code
            return

        module_name = node.module
        if not module_name:
            return

        for alias in node.names:
            yield common.ImportDecl(module_name, alias.name, None)


def parse_imports(tree: ast.Module) -> typ.Tuple[int, int, typ.Set[common.ImportDecl]]:
    future_imports_offset = 0
    imports_end_offset    = 0

    import_decls: typ.Set[common.ImportDecl] = set()

    for body_offset, node in enumerate(tree.body):
        is_docstring = (
            body_offset == 0 and isinstance(node, ast.Expr) and isinstance(node.value, ast.Str)
        )
        if is_docstring:
            future_imports_offset = body_offset + 1
            imports_end_offset    = body_offset + 1
            continue

        node_import_decls = list(find_import_decls(node))
        if not node_import_decls:
            # stop when we've passed the initial imports,
            # everything else is user code
            break

        for import_decl in node_import_decls:
            if import_decl.module_name == '__future__':
                future_imports_offset = body_offset
            imports_end_offset = body_offset
            import_decls.add(import_decl)

    return (future_imports_offset, imports_end_offset, import_decls)


def add_required_imports(tree: ast.Module, required_imports: typ.Set[common.ImportDecl]) -> None:
    """Add imports required by fixers.

    Some fixers depend on modules which may not be imported in
    the source module. As an example, occurrences of 'map' might
    be replaced with 'itertools.imap', in which case,
    "import itertools" will be added in the module scope.

    A further quirk is that all reqired imports must be added
    before any other statment. This is because that statement
    could be subject to the fix which requires the import. As
    a side effect, a module may end up being imported twice, if
    the module is imported after some statement.
    """
    (future_imports_offset, imports_end_offset, found_imports) = parse_imports(tree)

    missing_imports = sorted(required_imports - found_imports)

    import_node: ast.stmt
    for import_decl in missing_imports:
        if import_decl.import_name is None:
            import_node = ast.Import(names=[ast.alias(name=import_decl.module_name, asname=None)])
        else:
            import_node = ast.ImportFrom(
                module=import_decl.module_name,
                level=0,
                names=[ast.alias(name=import_decl.import_name, asname=None)],
            )

        if import_decl.py2_module_name:
            asname          = import_decl.import_name or import_decl.module_name
            fallback_import = ast.Import(
                names=[ast.alias(name=import_decl.py2_module_name, asname=asname)]
            )
            import_node = ast.Try(
                body=[import_node],
                handlers=[
                    ast.ExceptHandler(
                        type=ast.Name(id='ImportError', ctx=ast.Load()),
                        name=None,
                        body=[fallback_import],
                    )
                ],
                orelse=[],
                finalbody=[],
            )

        if import_decl.module_name == '__future__':
            tree.body.insert(future_imports_offset, import_node)
            future_imports_offset += 1
            imports_end_offset    += 1
        else:
            tree.body.insert(imports_end_offset, import_node)
            imports_end_offset += 1


def add_module_declarations(tree: ast.Module, module_declarations: typ.Set[str]) -> None:
    """Add global declarations required by fixers.

    Some fixers declare globals (or override builtins) the source
    module. As an example, occurrences of 'map' might be replaced
    by 'map = getattr(itertools, "map", map)'.

    These declarations are added directly after imports.
    """
    _, imports_end_offset, _ = parse_imports(tree)

    for decl_str in sorted(module_declarations):
        decl_node = utils.parse_stmt(decl_str)
        tree.body.insert(imports_end_offset + 1, decl_node)
        imports_end_offset += 1


def transpile_module(ctx: common.BuildContext, module_source: str) -> str:
    _module_header = module_source.split("import", 1)[0]
    _module_header = _module_header.split("'''", 1)[0]
    _module_header = _module_header.split('"""', 1)[0]

    lib3to6_mode_marker = MODE_MARKER_RE.search(_module_header)
    if lib3to6_mode_marker:
        mode = lib3to6_mode_marker.group('mode')
    else:
        mode = ctx.cfg.default_mode

    if mode == 'disabled':
        return module_source

    checker_names: FuzzyNames = ctx.cfg.checkers
    fixer_names  : FuzzyNames = ctx.cfg.fixers
    module_tree = ast.parse(module_source)
    required_imports   : typ.Set[common.ImportDecl] = set()
    module_declarations: typ.Set[str              ] = set()

    ver            = sys.version_info
    source_version = f"{ver.major}.{ver.minor}"
    target_version = ctx.cfg.target_version

    for checker in iter_fuzzy_selected_checkers(checker_names):
        if checker.version_info.is_applicable_to(source_version, target_version):
            checker(ctx, module_tree)

    for fixer in iter_fuzzy_selected_fixers(fixer_names):
        if fixer.version_info.is_applicable_to(source_version, target_version):
            maybe_fixed_module = fixer(ctx, module_tree)
            if maybe_fixed_module is None:
                raise Exception(f"Error running fixer {type(fixer).__name__}")
            required_imports.update(fixer.required_imports)
            module_declarations.update(fixer.module_declarations)
            module_tree = maybe_fixed_module

    if any(required_imports):
        add_required_imports(module_tree, required_imports)
    if any(module_declarations):
        add_module_declarations(module_tree, module_declarations)
    header = parse_module_header(module_source, target_version)
    return header.text + "".join(astor.to_source(module_tree))


def transpile_module_data(ctx: common.BuildContext, module_source_data: bytes) -> bytes:
    target_version      = ctx.cfg.target_version
    header              = parse_module_header(module_source_data, target_version)
    module_source       = module_source_data.decode(header.coding)
    fixed_module_source = transpile_module(ctx, module_source)
    return fixed_module_source.encode(header.coding)



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import ast
import typing as typ

from . import utils
from . import common
from . import checker_base as cb
from .checkers_backports import NoUnusableImportsChecker


class NoStarImports(cb.CheckerBase):
    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> None:
        for node in ast.walk(tree):
            if not isinstance(node, ast.ImportFrom):
                continue

            for alias in node.names:
                if alias.name == "*":
                    raise common.CheckError(f"Prohibited from {node.module} import *.", node)


def _iter_scope_names(tree: ast.Module) -> typ.Iterable[typ.Tuple[str, ast.AST]]:
    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.ClassDef)):
            yield node.name, node
        elif isinstance(node, ast.Name) and isinstance(node.ctx, ast.Store):
            yield node.id, node
        elif isinstance(node, (ast.ImportFrom, ast.Import)):
            for alias in node.names:
                name = alias.name if alias.asname is None else alias.asname
                yield name, node
        elif isinstance(node, ast.arg):
            yield node.arg, node


class NoOverriddenFixerImportsChecker(cb.CheckerBase):
    """Don't override names that fixers may reference."""

    prohibited_import_overrides = {"itertools", "six", "builtins"}

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> None:
        for name_in_scope, node in _iter_scope_names(tree):
            is_fixer_import = (
                isinstance(node, ast.Import)
                and len(node.names) == 1
                and node.names[0].asname is None
                and node.names[0].name == name_in_scope
            )
            if is_fixer_import:
                continue

            if name_in_scope in self.prohibited_import_overrides:
                msg = f"Prohibited override of import '{name_in_scope}'"
                raise common.CheckError(msg, node)


class NoOverriddenBuiltinsChecker(cb.CheckerBase):
    """Don't override names that fixers may reference."""

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> None:
        for name_in_scope, node in _iter_scope_names(tree):
            if name_in_scope in common.BUILTIN_NAMES:
                msg = f"Prohibited override of builtin '{name_in_scope}'"
                raise common.CheckError(msg, node)


PROHIBITED_OPEN_ARGUMENTS = {"encoding", "errors", "newline", "closefd", "opener"}


class NoOpenWithEncodingChecker(cb.CheckerBase):

    version_info = common.VersionInfo(apply_until="2.7")

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> None:
        for node in ast.walk(tree):
            if not isinstance(node, ast.Call):
                continue

            func_node = node.func
            if not isinstance(func_node, ast.Name):
                continue
            if func_node.id != "open" or not isinstance(func_node.ctx, ast.Load):
                continue

            mode = "r"
            if len(node.args) >= 2:
                mode_node = node.args[1]
                if not isinstance(mode_node, ast.Str):
                    msg = (
                        "Prohibited value for argument 'mode' of builtin.open. "
                        + f"Expected ast.Str node, got: {mode_node}"
                    )
                    raise common.CheckError(msg, node)

                mode = mode_node.s

            if len(node.args) > 3:
                raise common.CheckError("Prohibited positional arguments to builtin.open", node)

            for keyword in node.keywords:
                if keyword.arg in PROHIBITED_OPEN_ARGUMENTS:
                    msg = f"Prohibited keyword argument '{keyword.arg}' to builtin.open."
                    raise common.CheckError(msg, node)
                if keyword.arg != 'mode':
                    continue

                mode_node = keyword.value
                if not isinstance(mode_node, ast.Str):
                    msg = (
                        "Prohibited value for argument 'mode' of builtin.open. "
                        + f"Expected ast.Str node, got: {mode_node}"
                    )
                    raise common.CheckError(msg, node)

                mode = mode_node.s

            if "b" not in mode:
                msg = (
                    f"Prohibited value '{mode}' for argument 'mode' of builtin.open. "
                    + "Only binary modes are allowed, use io.open as an alternative."
                )
                raise common.CheckError(msg, node)


class NoAsyncAwait(cb.CheckerBase):

    version_info = common.VersionInfo(apply_until="3.4", works_since="3.5")

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> None:
        async_await_node_types = (ast.AsyncFor, ast.AsyncWith, ast.AsyncFunctionDef, ast.Await)
        for node in ast.walk(tree):
            if not isinstance(node, async_await_node_types):
                continue

            if isinstance(node, ast.AsyncFor):
                keywords = "async for"
            elif isinstance(node, ast.AsyncWith):
                keywords = "async with"
            elif isinstance(node, ast.AsyncFunctionDef):
                keywords = "async def"
            elif isinstance(node, ast.Await):
                keywords = "await"
            else:
                # probably dead codepath
                keywords = "async/await"

            msg = (
                f"Prohibited use of '{keywords}', which is not supported "
                f"for target_version='{ctx.cfg.target_version}'."
            )
            raise common.CheckError(msg, node)


class NoYieldFromChecker(cb.CheckerBase):

    version_info = common.VersionInfo(apply_until="3.2", works_since="3.3")

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> None:

        for node in ast.walk(tree):
            if isinstance(node, ast.YieldFrom):
                msg = (
                    "Prohibited use of 'yield from', which is not supported "
                    f"for your target_version={ctx.cfg.target_version}"
                )
                raise common.CheckError(msg, node)


class NoMatMultOpChecker(cb.CheckerBase):

    version_info = common.VersionInfo(apply_until="3.4", works_since="3.5")

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> None:
        if not hasattr(ast, 'MatMult'):
            return

        for node in ast.walk(tree):
            if not isinstance(node, ast.BinOp):
                continue

            if not isinstance(node.op, ast.MatMult):
                continue

            msg = "Prohibited use of matrix multiplication '@' operator."
            raise common.CheckError(msg, node)


def _raise_if_complex_named_tuple(node: ast.ClassDef) -> None:
    for subnode in node.body:
        if isinstance(subnode, ast.Expr) and isinstance(subnode.value, ast.Str):
            # docstring is fine
            continue

        if isinstance(subnode, ast.AnnAssign):
            if subnode.value:
                tgt = subnode.target
                assert isinstance(tgt, ast.Name)
                msg = (
                    "Prohibited use of default value "
                    + f"for field '{tgt.id}' of class '{node.name}'"
                )
                raise common.CheckError(msg, subnode, node)
        elif isinstance(subnode, ast.FunctionDef):
            msg = "Prohibited definition of method " + f"'{subnode.name}' for class '{node.name}'"
            raise common.CheckError(msg, subnode, node)
        else:
            msg = f"Unexpected subnode defined for class {node.name}: {subnode}"
            raise common.CheckError(msg, subnode, node)


class NoComplexNamedTuple(cb.CheckerBase):

    version_info = common.VersionInfo(apply_until="3.4", works_since="3.5")

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> None:
        _typing_module_name   : typ.Optional[str] = None
        _namedtuple_class_name: str = "NamedTuple"

        for node in ast.walk(tree):
            if isinstance(node, ast.Import):
                for alias in node.names:
                    if alias.name == 'typing':
                        _typing_module_name = alias.name if alias.asname is None else alias.asname
                continue

            if isinstance(node, ast.ImportFrom) and node.module == 'typing':
                for alias in node.names:
                    if alias.name == 'NamedTuple':
                        _namedtuple_class_name = (
                            alias.name if alias.asname is None else alias.asname
                        )
                continue

            is_namedtuple_class = (
                isinstance(node, ast.ClassDef)
                and (_typing_module_name or _namedtuple_class_name)
                and utils.has_base_class(node, _typing_module_name, _namedtuple_class_name)
            )
            if is_namedtuple_class:
                assert isinstance(node, ast.ClassDef), "mypy is stupid sometimes"
                _raise_if_complex_named_tuple(node)


# NOTE (mb 2018-06-24): I don't know how this could be done reliably.
#   The main issue is that there are objects other than dict, which
#   have methods named items,keys,values which this check wouldn't
#   apply to.
# class NoAssignedDictViews(cb.CheckerBase):
#
#     check_before = "3.0"
#
#     def __call__(self, ctx: common.BuildContext, tree: ast.Module):
#         pass

__all__ = [
    'NoStarImports',
    'NoOverriddenFixerImportsChecker',
    'NoOverriddenBuiltinsChecker',
    'NoOpenWithEncodingChecker',
    'NoAsyncAwait',
    'NoComplexNamedTuple',
    'NoUnusableImportsChecker',
    'NoYieldFromChecker',
    'NoMatMultOpChecker',
]



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2020 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
import ast

from . import common


class CheckerBase:

    # no info -> always apply
    version_info: common.VersionInfo = common.VersionInfo()

    def __call__(self, ctx: common.BuildContext, tree: ast.Module) -> None:
        raise NotImplementedError()



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import os
import sys
import setuptools


def project_path(*sub_paths):
    project_dirpath = os.path.abspath(os.path.dirname(__file__))
    return os.path.join(project_dirpath, *sub_paths)


def read(*sub_paths):
    with open(project_path(*sub_paths), mode="rb") as fobj:
        return fobj.read().decode("utf-8")


install_requires = [
    line.strip()
    for line in read("requirements", "pypi.txt").splitlines()
    if line.strip() and not line.startswith("#")
]


packages = setuptools.find_packages(project_path("src"))
package_dir = {"": "src"}


if any(arg.startswith("bdist") for arg in sys.argv):
    try:
        import lib3to6
        package_dir = lib3to6.fix(package_dir, install_requires=install_requires)
    except ImportError as ex:
        if "lib3to6" in str(ex):
            print("WARNING: 'lib3to6' missing, package will not be universal")
        else:
            raise


long_description = (read("README.md") + "\n\n" + read("CHANGELOG.md"))


setuptools.setup(
    name="lib3to6",
    license="MIT",
    author="Manuel Barkhau",
    author_email="mbarkhau@gmail.com",
    url="https://gitlab.com/mbarkhau/lib3to6",
    version="202009.1042",
    keywords="six lib2to3 astor ast",
    description="Compile Python 3.6+ code to Python 2.7+",
    long_description=long_description,
    long_description_content_type="text/markdown",

    packages=packages,
    package_dir=package_dir,
    install_requires=install_requires,
    entry_points="""
        [console_scripts]
        lib3to6=lib3to6.__main__:main
    """,
    python_requires=">=3.6",
    zip_safe=True,

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import ast
import typing as typ

import astor

from . import common
from . import transpile

# Recursive types not fully supported yet, nested types replaced with "Any"
# NodeOrNodelist = typ.Union[ast.AST, typ.List["NodeOrNodelist"]]
NodeOrNodelist = typ.Union[ast.AST, typ.List[typ.Any]]


# https://gist.github.com/marsam/d2a5af1563d129bb9482
def dump_ast(
    node              : typ.Any,
    annotate_fields   : bool = True,
    include_attributes: bool = False,
    indent            : str  = "  ",
) -> str:
    """Return a formatted dump of the tree in *node*.

    This is mainly useful for debugging purposes.  The returned
    string will show the names and the values for fields.  This
    makes the code impossible to evaluate, so if evaluation is
    wanted *annotate_fields* must be set to False.  Attributes
    such as line numbers and column offsets are not dumped by
    default.  If this is wanted, *include_attributes* can be set
    to True.
    """

    def _format(node: NodeOrNodelist, level: int = 1) -> str:
        # pylint: disable=protected-access
        if isinstance(node, ast.AST):
            fields = [(a, _format(b, level + 1)) for a, b in ast.iter_fields(node)]
            if include_attributes and node._attributes:
                fields.extend([(a, _format(getattr(node, a), level + 1)) for a in node._attributes])

            if annotate_fields:
                field_parts = ["%s=%s" % field for field in fields]
            else:
                field_parts = [b for a, b in fields]

            node_name     = node.__class__.__name__
            is_short_node = len(field_parts) <= 1 or isinstance(
                node, (ast.Name, ast.Num, ast.Str, ast.Bytes, ast.alias)
            )

            if is_short_node:
                return node_name + "(" + ", ".join(field_parts) + ")"

            lines = [node_name + "("]
            for part in field_parts:
                lines.append((indent * level) + part + ",")
            lines.append((indent * (level - 1)) + ")")
            return "\n".join(lines)
        elif isinstance(node, list):
            subnodes = node
            if len(subnodes) == 0:
                return "[]"

            if len(subnodes) == 1:
                return "[" + _format(subnodes[0], level) + "]"

            lines = [indent * level + _format(subnode, level + 1) + "," for subnode in subnodes]
            return "[\n" + "\n".join(lines) + "\n" + indent * (level - 1) + "]"
        return repr(node)

    if isinstance(node, (ast.AST, list)):
        return _format(node)
    else:
        raise TypeError("expected AST, got %r" % node.__class__.__name__)


def clean_whitespace(fixture_str: str) -> str:
    if fixture_str.strip().count("\n") == 0:
        return fixture_str.strip()

    fixture_lines = [line for line in fixture_str.splitlines() if line.strip()]
    line_indents  = [len(line) - len(line.lstrip()) for line in fixture_lines]
    if not any(line_indents) or min(line_indents) == 0:
        return fixture_str

    indent         = min(line_indents)
    dedented_lines = [line[indent:] for line in fixture_lines]
    return "\n".join(dedented_lines).strip() + "\n"


def parse_stmt(code: str) -> ast.stmt:
    module = ast.parse(code)
    assert len(module.body) == 1
    return module.body[0]


def parsedump_ast(code: str, mode: str = "exec", **kwargs) -> str:
    """Parse some code from a string and pretty-print it."""
    node = ast.parse(clean_whitespace(code), mode=mode)
    return dump_ast(node, **kwargs)


def parsedump_source(code: str, mode: str = "exec") -> str:
    node = ast.parse(clean_whitespace(code), mode=mode)
    return astor.to_source(node)


def transpile_and_dump(ctx: common.BuildContext, module_str: str) -> typ.Tuple[str, str, str]:
    module_str = clean_whitespace(module_str)
    header     = transpile.parse_module_header(module_str, ctx.cfg.target_version)
    result_str = transpile.transpile_module(ctx, module_str)
    return header.coding, header.text, result_str


def has_base_class(
    cls_node: ast.ClassDef, module_name: str = None, base_class_name: str = None
) -> bool:
    if not (module_name or base_class_name):
        return False

    for base in cls_node.bases:
        if isinstance(base, ast.Attribute):
            val = base.value
            if isinstance(val, ast.Name) and val.id == module_name and base.attr == base_class_name:
                return True

        if isinstance(base, ast.Name) and base.id == base_class_name:
            return True

    return False



# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

from .utils import parsedump_ast
from .utils import parsedump_source
from .packaging import fix
from .transpile import transpile_module

__version__ = "v202009.1042"

__all__ = ["fix", "transpile_module", "parsedump_ast", 'parsedump_source']
